# -*- coding: utf-8 -*-

# Copyright (c) 2006-2007, Rectorate of the University of Freiburg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the Freiburg Materials Research Center,
#   University of Freiburg nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

u"""
"""


import threading, math
import Queue
import wx
import sogl
import pyphant.core.CompositeWorker
import pyphant.core.Connectors as Connectors
import pyphant.core.WorkerRegistry
import pyphant.wxgui2.wxPyphantApplication
import pyphant.wxgui2.DataVisReg
from pyphant.core import Param
import cPickle as pickle

class PyphantDiagram(sogl.Diagram):
    def __init__(self, canvas, recipe=pyphant.core.CompositeWorker.CompositeWorker()):
        sogl.Diagram.__init__(self)
        self.SetCanvas(canvas)
        canvas.SetDiagram(self)
        self.workers={}
        self.connectors={}
        self.recipe = recipe
        self.createMainConnectors()
        self.recipe.registerListener(self.addWorkerShape, pyphant.core.CompositeWorker.WorkerAddedEvent)
        self.recipe.registerListener(self.onConnectionEvent, pyphant.core.CompositeWorker.ConnectionEvent)
        self.recipe.registerListener(self.onConnectorsExternalizationStateChangedEvent,
                                     pyphant.core.CompositeWorker.ConnectorsExternalizationStateChangedEvent)
        self.recipe.generateEvents({pyphant.core.CompositeWorker.WorkerAddedEvent:self.addWorkerShape,
                                    pyphant.core.CompositeWorker.ConnectionEvent: self.onConnectionEvent,
                                    pyphant.core.CompositeWorker.ConnectorsExternalizationStateChangedEvent:
                                    self.onConnectorsExternalizationStateChangedEvent})
        canvas.Bind(wx.EVT_SIZE, self.onResizeCanvas)

    def createMainConnectors(self):
        sockets = self.recipe.getSockets()
        plugs = self.recipe.getPlugs()
        canvas = self.GetCanvas()
        self.sockets = ConnectorBox(self, sockets, 'Plug')
        self.sockets.Move(canvas.getDC(), math.ceil(self.sockets.GetWidth()/2.0),
                          math.ceil(self.sockets.GetHeight()/2.0), True)
        self.plugs = ConnectorBox(self, plugs, 'Socket')

    def onResizeCanvas(self, event):
        canvas = self.GetCanvas()
        size = canvas.GetClientSize()
        self.plugs.Move(canvas.getDC(), size.x-math.ceil(self.plugs.GetWidth()/2.0),
                        size.y-math.ceil(self.plugs.GetHeight()/2.0), True)

    def onConnectorsExternalizationStateChangedEvent(self, event):
        if event.worker in self.workers:
            self.workers[event.worker].showConnector(event.connector)

    def onConnectionEvent(self, event):
        try:
            self.connectors[event.socket].updateIncomingPlugs()
        except KeyError:
            pass
        except AttributeError:
            pass

    def addWorkerShape(self, event):
        worker=event.worker
        canvas = self.GetCanvas()
        pos=worker.getAnnotation("pos")
        workerShape=WorkerShape(worker, self)
        workerShape.Move(canvas.getDC(),pos[0],pos[1])
        self.workers[worker]=workerShape
        canvas.Refresh()

    def addWorker(self, pickledWorkerInfo, pos=(0,0)):
        workerInfo=pickle.loads(str(pickledWorkerInfo))
        worker = workerInfo.createWorker(self.recipe, {"pos":pos})

class ConnectionShape(sogl.LineShape):
    def __init__(self, plug, socket):
        sogl.LineShape.__init__(self)
        self._plug = plug
        self._socket = socket
        canvas = self._socket.GetCanvas()
        self.SetCanvas(canvas)
        self.AddArrow(sogl.ARROW_ARROW)
        self.Show(True)
        canvas.GetDiagram().AddShape(self)
        self.MakeLineControlPoints(2)
        self._plug.AddLine(self, self._socket)

    def destroy(self):
        self._plug.RemoveLine(self)
        canvas = self.GetCanvas()
        canvas.GetDiagram().RemoveShape(self)
        self._socket.connector.pullPlug()
        canvas.Refresh()

    def OnRightClick(self,x,y,keys,attachment):
        self.destroy()

class TypeColors(dict):
    def __init__(self):
        dict.__init__(self)
        self[Connectors.TYPE_ARRAY]=wx.BLUE_BRUSH
        self[Connectors.TYPE_IMAGE]=wx.RED_BRUSH
        self[Connectors.TYPE_INACTIVE]=wx.GREY_BRUSH
        self[Connectors.TYPE_INT]=wx.GREEN_BRUSH
        self[Connectors.TYPE_STRING]=wx.Brush("#FFFF00")
        self[Connectors.TYPE_BOOL]=wx.Brush("#FF00FF")
        self._unknownType=wx.BLACK_BRUSH

    def __getitem__(self, key):
        return self.setdefault(key, self._unknownType)

class ConnectorShape(object):
    TYPE_COLORS=TypeColors()
    def __init__(self, connector):
        self.SetBrush(self.TYPE_COLORS[connector.type])
        self.connector=connector

    def OnBeginDragLeft(self, x, y, keys, attachments):
        pass

    def OnDragLeft(self, draw, x, y, keys, attachments):
        pass

    #necessary to keep ogl from doing stupid stuff
    def OnEndDragLeft(self, x, y, keys, attachments):
        pass


class AcceptPlugs(object):
    def updateIncomingPlugs(self):
        p=self.connector.getPlug()
        if p:
            plugShape=self.GetCanvas().GetDiagram().connectors[p]
            ConnectionShape(plugShape, self)

class SocketShape(ConnectorShape, sogl.RectangleShape, AcceptPlugs):
    def __init__(self, socket):
        sogl.RectangleShape.__init__(self, 10, 10)
        ConnectorShape.__init__(self, socket)

class ParamShape(ConnectorShape, sogl.PolygonShape, AcceptPlugs):
    def __init__(self, param):
        sogl.PolygonShape.__init__(self)
        ConnectorShape.__init__(self, param)
        self._param=param
        r=5
        self.Create([(r,0),(0,r),(-r,0),(0,-r)])

    def updateIncomingPlugs(self):
        super(ParamShape, self).updateIncomingPlugs()
        if self.connector.getPlug():
            self.SetBrush(self.TYPE_COLORS[self.connector.type])
        else:
            self.SetBrush(self.TYPE_COLORS[Connectors.TYPE_INACTIVE])
        self.Draw(self.GetCanvas().getDC())

class ProgressMeter(wx.ProgressDialog):
    def __init__(self, plugName):
        title = u"Calculating %s" % plugName
        message = u"Calculating the result of %s" % plugName
        wx.ProgressDialog.__init__(self, title, message, parent=wx.GetApp().getMainFrame())
        self.processes={}
        self.count=0
        self.cv=threading.Condition()
        self.percentage=0
        self.share=1.0

    def update(self):
        self.cv.acquire()
        self.cv.wait(0.1)
        self.Update(self.percentage)
        self.cv.release()

    def startProcess(self, process):
        self.processes[process]=0
        self.share=1.0/len(self.processes)
        self.updatePercentage()

    def updatePercentage(self):
        self.cv.acquire()
        self.percentage=int(sum([x*self.share for x in self.processes.values()]))
        self.count+=1
        self.cv.notify()
        self.cv.release()

    def updateProcess(self, process, value):
        self.processes[process]=value
        self.updatePercentage()

    def finishProcess(self, process):
        self.processes[process]=100
        self.updatePercentage()

    def finishAllProcesses(self):
        self.cv.acquire()
        self.percentage=100
        self.updatePercentage()
        self.cv.release()


class PlugShape(ConnectorShape, sogl.CircleShape):
    dataVisReg = pyphant.wxgui2.DataVisReg.DataVisReg.getInstance()
    def __init__(self, plug):
        sogl.CircleShape.__init__(self, 10)
        ConnectorShape.__init__(self, plug)
        self._plug=plug

    def _createContextMenu(self):
        self._ids={}
        self._menu = wx.Menu()
        visMenu = wx.Menu()
        for visualizer in self.dataVisReg.getVisualizers(self._plug.type):
            id=wx.NewId()
            self._ids[id]=visualizer
            visMenu.Append(id, visualizer.name)
            self.GetCanvas().Bind(wx.EVT_MENU, self.visualize, id=id)
        self._menu.AppendMenu(-1, "Visualize", visMenu)

    def visualize(self, event):
        if not self._plug.resultIsAvailable():
            progress = ProgressMeter(self._plug.name)
            exception_queue = Queue.Queue()
            computer = Connectors.Computer(self._plug.getResult, exception_queue, subscriber=progress)
            computer.start()
            while (progress.percentage<100) and (computer.isAlive()):
                progress.update()
            progress.finishAllProcesses()
            progress.Destroy()
            if computer.isAlive():
                computer.join()
            result = computer.result
        else:
            result = self._plug._result
        if not result == None:
            self._ids[event.GetId()](result)
        else:
            wx.MessageBox(u"No result is available.\nPlease check the logs.",
                          u"No Result", wx.ICON_ERROR)

    def OnRightClick(self, x, y, keys, attachments):
        if not hasattr(self, '_menu'):
            self._createContextMenu()
        self.GetCanvas().PopupMenu(self._menu)

    def OnEndDragLeft(self, x, y, keys, attachments):
        nearest_object, attachment = self.GetCanvas().FindShape(x,y)
        if isinstance(nearest_object, AcceptPlugs):
            nearest_object.connector.insert(self._plug)
            #nearest_object.updateIncomingPlugs()


class BodyShape(sogl.RectangleShape):
    def __init__(self, workerShape, name):
        sogl.RectangleShape.__init__(self,100,50)
        self.workerShape=workerShape
        self.RemoveSensitivityFilter(sogl.OP_DRAG_LEFT|sogl.OP_DRAG_RIGHT)
        self.SetText(name)

    def SetText(self, string):
        """Add a line of text to the shape's default text region."""
        if not self._regions:
            return
        region = self._regions[0]
        region.ClearText()
        new_line = sogl.ShapeTextLine(0, 0, string)
        text = region.GetFormattedText()
        text.append(new_line)
        self._formatted = False

    def OnRightClick(self, x, y, keys, attachment):
        frame=wx.GetTopLevelParent(self.GetCanvas())
        frame._wxPyphantApp.configureWorker(self.workerShape.worker)

    def OnLeftClick(self, x, y, keys, attachment):
        frame=wx.GetTopLevelParent(self.GetCanvas())
        frame._wxPyphantApp.configureWorker(self.workerShape.worker)
        #frame._wxPyphantApp.editCompositeWorker(self.workerShape.worker)


class ConnectorBox(sogl.CompositeShape):
    def __init__(self, diag, connectors, conType):
        sogl.CompositeShape.__init__(self)
        self.diag = diag
        self.SetCanvas(diag.GetCanvas())
        self.connectors = connectors
        self.conShapes = []
        for c in self.connectors:
            if conType == 'Plug':
                cs = PlugShape(c)
            elif conType == 'Socket':
                cs = SocketShape(c)
            else:
                pass
            self.diag.connectors[c] = cs
            self.conShapes.append(cs)
            self.AddChild(cs)
        self.AddConstraint(sogl.Constraint(sogl.CONSTRAINT_ALIGNED_TOP,
                                           self,self.conShapes))
        if len(self.conShapes)>0:
            self.AddConstraint(sogl.Constraint(sogl.CONSTRAINT_ALIGNED_LEFT,
                                               self,[self.conShapes[0]]))
        for i in xrange(1, len(self.conShapes)):
            self.AddConstraint(sogl.Constraint(sogl.CONSTRAINT_RIGHT_OF,
                                               self.conShapes[i-1],
                                               [self.conShapes[i]]))
        self.Recompute()
        self.Show(True)

class WorkerShape(sogl.CompositeShape):
    def __init__(self, worker, diag):
        sogl.CompositeShape.__init__(self)
        self.diag = diag
        self.SetCanvas(diag.GetCanvas())
        self.worker=worker
        bodyShape = BodyShape(self,worker.getParam('name').value)
        self.AddChild(bodyShape)

        self.socketShapes=[]
        for socket in worker.getSockets():
            socketShape=SocketShape(socket)
            self.diag.connectors[socket]=socketShape
            self.AddChild(socketShape)
            self.socketShapes.append(socketShape)

        self.plugShapes=[]
        for plug in worker.getPlugs():
            plugShape=PlugShape(plug)
            self.diag.connectors[plug]=plugShape
            self.AddChild(plugShape)
            self.plugShapes.append(plugShape)

        self.paramShapes=[]
        for param in worker.getParamList():
            paramShape=ParamShape(param)
            self.diag.connectors[param]=paramShape
            self.AddChild(paramShape)
            self.paramShapes.append(paramShape)

        if self.socketShapes != []:
            self.AddConstraint(sogl.Constraint(sogl.CONSTRAINT_ALIGNED_TOP,bodyShape,
                                              self.socketShapes))
            self.AddConstraint(sogl.Constraint(sogl.CONSTRAINT_ALIGNED_LEFT,bodyShape,
                                              [self.socketShapes[0]]))
            for i in xrange(1,len(self.socketShapes)):
                self.AddConstraint(sogl.Constraint(sogl.CONSTRAINT_RIGHT_OF, self.socketShapes[i-1],
                                                  [self.socketShapes[i]]))

        if self.plugShapes != []:
            self.AddConstraint(sogl.Constraint(sogl.CONSTRAINT_ALIGNED_BOTTOM,bodyShape,
                                              self.plugShapes))
            self.AddConstraint(sogl.Constraint(sogl.CONSTRAINT_ALIGNED_RIGHT,bodyShape,
                                              [self.plugShapes[0]]))
            for i in xrange(1,len(self.plugShapes)):
                self.AddConstraint(sogl.Constraint(sogl.CONSTRAINT_LEFT_OF, self.plugShapes[i-1],
                                                  [self.plugShapes[i]]))

        if self.paramShapes != []:
            self.AddConstraint(sogl.Constraint(sogl.CONSTRAINT_ALIGNED_TOP,bodyShape,
                                              self.paramShapes))
            self.AddConstraint(sogl.Constraint(sogl.CONSTRAINT_ALIGNED_RIGHT,bodyShape,
                                              [self.paramShapes[0]]))
            for i in xrange(1,len(self.paramShapes)):
                self.AddConstraint(sogl.Constraint(sogl.CONSTRAINT_LEFT_OF,self.paramShapes[i-1],
                                                  [self.paramShapes[i]]))
        self.Recompute()

        self.RemoveSensitivityFilter(sogl.OP_CLICK_LEFT, False)
        #self.AddSensitivityFilter(sogl.OP_CLICK_LEFT, True)

        self.worker.registerParamListener(lambda event: (bodyShape.SetText(event.newValue), bodyShape.GetCanvas().Refresh()), 'name', Param.ParamChanged)
        self.Show(True)

        map(lambda con: self.diag.connectors[con].Show(False),
            filter(lambda connector: not connector.isExternal, self.diag.connectors))

    def hideConnector(self, connector):
        try:
            self.diag.connectors[connector].Show(False)
        except KeyError, e:
            pass

    def showConnector(self, connector):
        try:
            self.diag.connectors[connector].Show(True)
        except KeyError, e:
            pass

    def OnMovePost(self, dc, x, y, old_x, old_y, display = True):
        self.worker._annotations["pos"]=(x,y)




from pyphant.quantities import Quantity
from ast import (NodeTransformer, fix_missing_locations, Name, Load,
                 BinOp, Num, Mult, Compare, BoolOp, And, Add, Sub, Div, Or,
                 Call, Not, Expression)


class LocationFixingNodeTransformer(NodeTransformer):
    def visit(self, *args, **kargs):
        result = NodeTransformer.visit(self, *args, **kargs)
        fix_missing_locations(result)
        return result


class ReplaceName(LocationFixingNodeTransformer):
    def __init__(self, sampleContainer):
        self.localDict = {}
        self.count = 0
        self.sc = sampleContainer

    def visit_Call(self, node):
        if isinstance(node.func, Name) and node.func.id.lower() == 'col':
            newName = self.getName(self.sc[node.args[0].s])
            return Name(newName, Load())

    def visit_Str(self, node):
        quantity = Quantity(node.s)
        class QuantityDummy(object):
            pass
        dummy = QuantityDummy()
        dummy.unit = Quantity(1.0, quantity.unit)
        dummy.data = quantity.value
        dummy.dimensions = None
        newName = self.getName(dummy)
        return Name(newName, Load())

    def getName(self, ref):
        newName = "N%s" % self.count
        self.count += 1
        self.localDict[newName] = ref
        return newName


def withFactor(factor, node):
    if not isinstance(factor, float):
        raise ValueError('Incompatible units!')
    if factor == 1.0:
        return node
    return BinOp(Num(factor), Mult(), node)


class ReplaceCompare(LocationFixingNodeTransformer):
    def __init__(self, localDict):
        self.localDict = localDict

    def visit_Compare(self, node):
        self.generic_visit(node)
        unitcalc = UnitCalculator(self.localDict)
        leftUD = unitcalc.getUnitAndDim(node.left)
        listUD = [unitcalc.getUnitAndDim(comp) for comp in node.comparators]
        nonNoneDims = [ud[1] for ud in listUD + [leftUD] if ud[1] is not None]
        for dims in nonNoneDims:
            checkDimensions(nonNoneDims[0], dims)
        factorlist = [ud[0] / leftUD[0] for ud in listUD]
        newComplist = [withFactor(*t) \
                       for t in zip(factorlist, node.comparators)]
        compOp = Compare(node.left, node.ops, newComplist)
        compOpTrans = self.compBreaker(compOp)
        return compOpTrans

    def compBreaker(self, node):
        assert isinstance(node, Compare)
        if len(node.comparators) == 1:
            return node
        else:
            comp1 = Compare(node.left, node.ops[0:1],
                            node.comparators[0:1])
            comp2 = Compare(node.comparators[0],
                            node.ops[1:], node.comparators[1:])
            newNode = BoolOp(And(), [comp1, self.compBreaker(comp2)])
            return newNode


class ReplaceOperator(LocationFixingNodeTransformer):
    def __init__(self, localDict):
        self.localDict = localDict

    def visit_BinOp(self, node):
        self.generic_visit(node)
        unitcalc = UnitCalculator(self.localDict)
        leftUD = unitcalc.getUnitAndDim(node.left)
        rightUD = unitcalc.getUnitAndDim(node.right)
        checkDimensions(leftUD[1], rightUD[1])
        if isinstance(node.op, (Add, Sub)):
            factor = rightUD[0] / leftUD[0]
            right = withFactor(factor, node.right)
            binOp = BinOp(node.left, node.op, right)
            return binOp
        elif isinstance(node.op, (Mult, Div)):
            return node
        else:
            raise NotImplementedError('%s not implemented' % (node.op, ))

    def visit_BoolOp(self, node):
        self.generic_visit(node)
        if isinstance(node.op, And):
            func = 'logical_and'
        elif isinstance(node.op, Or):
            func = 'logical_or'
        else:
            raise NotImplementedError('%s not implemented' % (node.op, ))
        return self.boolOpBreaker(node.values, func)

    def boolOpBreaker(self, values, func):
        if len(values) == 1:
            return values[0]
        else:
            return Call(Name(func, Load()),
                        [values[0], self.boolOpBreaker(values[1:], func)],
                        [], None, None)

    def visit_UnaryOp(self, node):
        self.generic_visit(node)
        if isinstance(node.op, Not):
            return Call(Name('logical_not', Load()), [node.operand],
                        [], None, None)
        else:
            raise NotImplementedError('%s not implemented' % (node.op, ))


class UnitCalculator(object):
    def __init__(self, localDict):
        self.localDict = localDict

    def getUnitAndDim(self, node):
        if isinstance(node, Expression):
            return self.getUnitAndDim(node.body)
        elif isinstance(node, Name):
            if node.id in ['True', 'False']:
                return (1.0, None)
            else:
                column = self.localDict[node.id]
                return (column.unit, column.dimensions)
        elif isinstance(node, Num):
            return (1.0, None)
        elif isinstance(node, Call):
            if not isinstance(node.func, Name):
                raise NotImplementedError(
                    'Dynamic functions are not implemented!')
            funcId = node.func.id
            if funcId in ['logical_and', 'logical_or']:
                left = self.getUnitAndDim(node.args[0])
                right = self.getUnitAndDim(node.args[1])
                dimensions = checkDimensions(left[1], right[1])
                if not isinstance(left[0], float):
                    raise ValueError(
                        "Type %s cannot be interpreted as a Boolean" % left)
                if not isinstance(right[0], float):
                    raise ValueError(
                        "Type %s cannot be interpreted as a Boolean" % right)
                return (1.0, dimensions)
            elif funcId == 'logical_not':
                return self.getUnitAndDim(node.args[0])
            else:
                raise NotImplementedError("Function '%s' not implemented" \
                                          % (funcId, ))
        elif isinstance(node, BinOp):
            left = self.getUnitAndDim(node.left)
            right = self.getUnitAndDim(node.right)
            dimensions = checkDimensions(left[1], right[1])
            if isinstance(node.op, (Add, Sub)):
                if not isinstance(left[0] / right[0], float):
                    raise ValueError("units %s, %s not compatible" \
                                     % (left, right))
                unit = left[0]
            elif isinstance(node.op, Mult):
                unit = left[0] * right[0]
            elif isinstance(node.op, Div):
                unit = left[0] / right[0]
            else:
                raise NotImplementedError()
            return (unit, dimensions)
        elif isinstance(node, Compare):
            left = self.getUnitAndDim(node.left)
            nonNoneDims = []
            if left[1] is not None:
                nonNoneDims.append(left[1])
            for comparator in node.comparators:
                right = self.getUnitAndDim(comparator)
                if right[1] is not None:
                    nonNoneDims.append(right[1])
                if not isinstance(left[0] / right[0], float):
                    raise ValueError("units %s, %s not compatible" \
                                     % (left[0], right[0]))
            if len(nonNoneDims) >= 1:
                for dims in nonNoneDims:
                    checkDimensions(nonNoneDims[0], dims)
                dimensions = nonNoneDims[0]
            else:
                dimensions = None
            return (1.0, dimensions)
        else:
            raise NotImplementedError()


def checkDimensions(dimensions1, dimensions2):
    if dimensions1 is not None and dimensions2 is not None and \
           dimensions1 != dimensions2:
        msg = 'Dimensions "%s" and "%s" do not match!' \
              % (dimensions1, dimensions2)
        raise ValueError(msg)
    return dimensions1 or dimensions2



# -*- coding: utf-8 -*-

# Copyright (c) 2006-2008, Rectorate of the University of Freiburg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the Freiburg Materials Research Center,
#   University of Freiburg nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import with_statement

"""
This module provides the KnowledgeManager class as well as some helper
classes.
"""


from pyphant.core.singletonmixin import Singleton
import tempfile
import os
import logging
import re
from pyphant.core.H5FileHandler import (H5FileHandler, im_id)
from pyphant.core import LoadFMF
from pyphant.core.SQLiteWrapper import (SQLiteWrapper, AnyValue)
from pyphant.core.Helpers import getPyphantPath
from uuid import uuid1
from urlparse import urlparse
import urllib

# Limit for sum(DC.rawDataBytes) for DC in cache:
CACHE_MAX_SIZE = 256 * 1024 * 1024
# Limit for number of stored DCs in cache:
CACHE_MAX_NUMBER = 100
KM_PATH = 'KMstorage'
REHDF5 = re.compile(r'..*\.h5$|..*\.hdf$|..*\.hdf5$')
REFMF = re.compile(r'..*\.fmf$')

def getFilenameFromDcId(dcId, temporary=False):
    """
    Returns a unique filename for the given emd5.
    """
    emd5list = (dcId + '.h5')[7:].split('/')
    emd5path = os.path.join(
        *(emd5list[:-2] + [emd5list[-2][:10],
                           emd5list[-2][11:] + '.' + emd5list[-1]]))
    #needed for windows version of pytables:
    emd5path = emd5path.replace(':', '.')
    directory = os.path.dirname(emd5path)
    filename = os.path.basename(emd5path)
    if temporary:
        subdir = os.path.join('tmp', 'by_emd5')
    else:
        subdir = 'by_emd5'
    return os.path.join(getPyphantPath(
        os.path.join(KM_PATH, subdir, directory)), filename)


class DCNotFoundError(Exception):
    pass


class CachedDC(object):
    """
    Class representing a cached DataContainer
    """
    def __init__(self, dc_ref):
        self.id = dc_ref.id
        self.ref = dc_ref
        self.size = dc_ref.rawDataBytes

    def __eq__(self, other):
        return self.id == other.id


class TestCachedDC(object):
    """
    Class representing a cache lookup
    """
    def __init__(self, dc_id):
        self.id = dc_id

    def __eq__(self, other):
        return self.id == other.id


KM_DBASE = u'default' # modify for debug purposes


class KnowledgeManager(Singleton):
    """
    Knowledge Manager for Pyphant
    =============================
    The ID of a DataContainer object is given by an emd5 string.
    Responsibilities:
    -----------------
    - Manage local storage of DataContainers
    - Resolve IDs to DC instances
    - Communicate with a KnowledgeNode in order to share Knowledge
      among other KM instances via HTTP.
    - Manage and search meta data for DataContainers
    Usage examples:
    ---------------
    Get a reference to the KnowledgeManager instance, which is a
    singleton:
    from pyphant.core.KnowledgeManager import KnowledgeManager
        km = KnowledgeManager.getInstance()
    Register a local HDF5 file:
        km.registerURL("/some_directory/data.h5")
    Register and persist a local HDF5 file:
        km.registerURL("file:///tmp/data.h5")
    Register and persist a remote FMF file:
        km.registerURL("http://example.com/repository/data.fmf")
    Request DataContainer by its id:
        dc = km.getDataContainer(id) # `dc` is an actual instance now
    For searching meta data see docstring of KM.search() below.
    How to share Knowledge:
        Hook up the KM to a KnowledgeNode, see documentation in the
        KnowledgeNode module.
    Known issues:
    ------------
    KM is NOT thread-safe yet.
    """
    def __init__(self):
        """
        Sets up the DataBase if it has not been initialized yet,
        sets up the cache and clears the tmp dir.
        Sets a uuid to identify the instance.
        """
        super(KnowledgeManager, self).__init__()
        self.logger = logging.getLogger("pyphant")
        self._cache = []
        self._cache_size = 0
        if KM_DBASE == u'default':
            self.dbase = os.path.join(getPyphantPath('sqlite3'),
                                      "km_meta.sqlite3")
        else:
            self.dbase = KM_DBASE
        self.any_value = AnyValue()
        self.node = None # for hooking up a KnowledgeNode
        self.uuid = uuid1().urn
        tmpdir = getPyphantPath(os.path.join(KM_PATH, 'tmp'))
        if os.path.isdir(tmpdir):
            from shutil import rmtree
            try:
                rmtree(tmpdir)
            except OSError:
                self.logger.warn("Could not delete '%s'." % tmpdir)
        with SQLiteWrapper(self.dbase) as wrapper:
            rebuild = wrapper.dbase_broken()
        if rebuild:
            self.logger.info("dbase needs rebuild")
            self.rebuildIndex()
        else:
            with SQLiteWrapper(self.dbase) as wrapper:
                wrapper.setup_sqlite()

    def rebuildIndex(self):
        self.logger.info("rebuilding dbase...")
        oldname = self.dbase + ".bak"
        if os.path.exists(oldname):
            os.remove(oldname)
        os.rename(self.dbase, oldname)
        with SQLiteWrapper(self.dbase) as wrapper:
            wrapper.setup_dbase()
        self.updateIndex()
        self.logger.info("done rebuilding dbase")

    def updateIndex(self):
        file_list = []
        def accumulate_files(flist, directory, files):
            for fname in [afname for afname in files if afname.endswith('.h5')]:
                flist.append(os.path.realpath(os.path.join(directory, fname)))
        os.path.walk(getPyphantPath(KM_PATH), accumulate_files, file_list)
        try:
            from wx import (ProgressDialog, PyNoAppError)
            try:
                pdial = ProgressDialog('Rebuilding index...', ' ' * 100,
                                       maximum=len(file_list))
            except PyNoAppError:
                pdial = None
        except ImportError:
            pdial = None
        count = 1
        for realname in file_list:
            if pdial is not None:
                pdial.Update(count, os.path.basename(realname))
                count += 1
            try:
                self.registerH5(realname)
            except:
                self.logger.warn("Could not extract meta data from '%s'."\
                                 % realname)
        if pdial is not None:
            pdial.Destroy()

    def hasDataContainer(self, dcid):
        """
        Returns whether the given DC is stored locally.
        """
        with SQLiteWrapper(self.dbase) as wrapper:
            has_entry = wrapper.has_entry(dcid)
        return has_entry

    def isTemporary(self, dcid):
        """
        Returns whether the given DC is stored temporarily.
        """
        with SQLiteWrapper(self.dbase) as wrapper:
            is_tmp = wrapper.is_temporary(dcid)
        return is_tmp

    def setTemporary(self, dcid, temporary):
        """
        Sets the given entry to temporary, which means it will be
        deleted upon restart or removes the temporary flag.
        - dcid: emd5 of DataContainer
        - temporary: boolean
        """
        with SQLiteWrapper(self.dbase) as wrapper:
            wrapper.set_temporary(dcid, temporary)

    def getH5FileHandler(self, filename, mode='r'):
        """
        Returns an H5FileHandler for the given filename to perform IO
        operations on the file in a save way.
        filename -- path to the HDF5 file
        mode -- see H5FileHandler
        """
        return H5FileHandler(filename, mode)

    def registerH5(self, filename, temporary=False):
        """
        Adds the given file to the knowledge pool. If you want the data to
        be copied to the .pyphant directory, use registerURL() instead.
        filename -- path to the HDF5 file
        temporary -- flag that marks data to be deleted upon next
                     instantiation of a KM Singleton
        """
        h5fh = self.getH5FileHandler(filename)
        with h5fh:
            summaryDict = h5fh.loadSummary()
        with SQLiteWrapper(self.dbase) as wrapper:
            for dcId, summary in summaryDict.items():
                if dcId == im_id:
                    wrapper.set_entry(summary, None, temporary)
                else:
                    wrapper.set_entry(summary, filename, temporary)

    def registerURL(self, url, temporary=False):
        """
        Registers an HDF5 or FMF file downloadable from given URL and stores it
        in the .pyphant directory. The content of the file is made
        available to the KnowledgeManager.
        HTTP redirects are resolved. The filetype is determined by the
        extension.
        url -- URL of the HDF5 or FMF file
        temporary -- set to True in order to mark the data to be deleted upon
                     next instantiation of a KM singleton
        """
        parsed = urlparse(url)
        tmp_extension = ''
        if temporary:
            tmp_extension = 'tmp'
        try:
            remote_path = parsed.path
        except AttributeError:
            remote_path = parsed[2]
        directory = os.path.join(KM_PATH, tmp_extension,
                                 'registered')
        filename = os.path.join(getPyphantPath(directory),
                                os.path.basename(remote_path))
        if os.path.exists(filename):
            i = 0
            directory = os.path.dirname(filename)
            basename = os.path.basename(filename)
            split = basename.split('.')
            ext = split.pop()
            fnwoext = ''
            for part in split:
                fnwoext += (part + '.')
            from sys import maxint
            while i < maxint:
                fill = str(i).zfill(10)
                tryfn = os.path.join(directory,
                                     "%s%s.%s" % (fnwoext, fill, ext))
                if os.path.exists(tryfn):
                    i += 1
                else:
                    filename = tryfn
                    break
        self.logger.info("Retrieving url '%s'..." % url)
        self.logger.info("Using local file '%s'." % filename)
        savedto, headers = urllib.urlretrieve(url, filename)
        if REFMF.match(filename.lower()) != None:
            self.registerFMF(filename, temporary)
        elif REHDF5.match(filename.lower()) != None:
            self.registerH5(filename, temporary)
        else:
            msg = "Could not guess type of '%s'" % url
            self.logger.error(msg)
            raise ValueError(msg)

    def registerDataContainer(self, dc, temporary=False):
        """
        Registers a DataContainer located in memory using a given
        reference and stores it in the pyphant directory.
        The DataContainer must have an .id attribute,
        which could be generated by the datacontainer.seal() method.
        If the DCs emd5 is already known to the KnowledgeManager,
        the DC is not registered again since emd5s are unique.
        dc -- reference to the DataContainer object
        temporary -- dc is stored only until another KM singleton is
                     created. Set this flag to True e.g. for unit tests
                     or whenever you do not want to produce garbage on
                     your hard drive.
        """
        if dc.id == None:
            msg = "Missing id for DataContainer. DC has not been sealed."
            self.logger.error(msg)
            raise ValueError(msg)
        if not self.hasDataContainer(dc.id):
            filename = getFilenameFromDcId(dc.id, temporary)
            handler = self.getH5FileHandler(filename, 'w')
            with handler:
                handler.saveDataContainer(dc)
            self.registerH5(filename, temporary)

    def registerFMF(self, filename, temporary=False):
        """
        Extracts a SampleContainer from a given FMF file and stores it
        permanently. The emd5 of the SampleContainer that has been generated
        is returned.
        filename -- path to the FMF file
        temporary -- see registerDataContainer
        """
        sc = LoadFMF.loadFMFFromFile(filename)
        self.registerDataContainer(sc, temporary)
        return sc.id

    def getDCFromCache(self, dc_id, filename):
        """
        Returns a DC instance from cache or local storage.
        Also puts DC to cache if reasonable.
        fc_id: emd5 to look for in cache
        filename: alternative source if dc_id not present in cache
        """
        try:
            index = self._cache.index(TestCachedDC(dc_id))
            cached = self._cache.pop(index)
            self._cache.append(cached)
            return cached.ref
        except ValueError:
            with self.getH5FileHandler(filename) as handler:
                dc = handler.loadDataContainer(dc_id)
            self._attemptToCacheDC(dc)
            return dc

    def _attemptToCacheDC(self, dc):
        cache_item = CachedDC(dc)
        if cache_item.size > CACHE_MAX_SIZE:
            return
        number_fits = len(self._cache) < CACHE_MAX_NUMBER
        self._cache.reverse()
        if not number_fits:
            self._cache_size -= self._cache.pop().size
        desired_size = CACHE_MAX_SIZE - cache_item.size
        not_size_fits = self._cache_size > desired_size
        while not_size_fits:
            self._cache_size -= self._cache.pop().size
            not_size_fits = self._cache_size > desired_size
        self._cache.reverse()
        self._cache.append(cache_item)
        self._cache_size += cache_item.size

    def getDataContainer(self, dc_id, use_cache=True, try_remote=True):
        """
        Returns DataContainer matching the given id.
        dc_id -- Unique ID of the DataContainer (emd5)
        use_cache -- Try local cache first and cache DC for further
                     lookups (default: True)
        try_remote -- Try to get DC from remote KMs (default: True)
        """
        filename = None
        with SQLiteWrapper(self.dbase) as wrapper:
            try:
                filename = wrapper[dc_id]['storage']
            except KeyError:
                pass
        if filename != None:
            if use_cache:
                return self.getDCFromCache(dc_id, filename)
            with self.getH5FileHandler(filename) as handler:
                dc = handler.loadDataContainer(dc_id)
            return dc
        elif try_remote and self.node != None:
            try:
                return self.node.get_datacontainer(dc_id)
            except DCNotFoundError:
                pass
        msg = "Could not find DC with id '%s'." % dc_id
        self.logger.error(msg)
        raise DCNotFoundError(msg)

    def getEmd5List(self):
        """
        returns a list with all locally known DataContainer ids.
        """
        with SQLiteWrapper(self.dbase) as wrapper:
            return wrapper.get_emd5_list()

    def search(self, result_keys, search_dict={}, order_by=None,
               order_asc=True, limit=-1, offset=0, distinct=False):
        """
        see SQLiteWrapper.get_andsearch_result
        """
        with SQLiteWrapper(self.dbase) as wrapper:
            return wrapper.get_andsearch_result(
                result_keys, search_dict, order_by, order_asc,
                limit, offset, distinct)

    def getSummary(self, dc_id):
        """
        This method returns a dictionary with meta information about
        the given DC.
        """
        with SQLiteWrapper(self.dbase) as wrapper:
            rowwrapper = wrapper[dc_id]
            keys = list(SQLiteWrapper.all_keys)
            if dc_id.endswith('field'):
                keys.remove('columns')
            elif dc_id.endswith('sample'):
                keys.remove('unit')
                keys.remove('dimensions')
            summary = dict([(key, rowwrapper[key]) for key in keys])
        return summary



# -*- coding: utf-8 -*-

# Copyright (c) 2006-2009, Rectorate of the University of Freiburg
# Copyright (c) 2009, Andreas W. Liehr (liehr@users.sourceforge.net)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the Freiburg Materials Research Center,
#   University of Freiburg nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import with_statement

u"""Provides unittest classes for H5FileHandler.
"""


import unittest
from pyphant.quantities import Quantity as PQ
from pyphant.core.DataContainer import FieldContainer, SampleContainer
from pyphant.core.H5FileHandler import H5FileHandler as H5FH
from numpy import array as NPArray
import os
from tempfile import mkstemp


class BasicTestCase(unittest.TestCase):
    def testReadOnlyFileNotFound(self):
        try:
            H5FH('', 'r')
            assert False
        except IOError:
            pass


class FieldContainerTestCase(unittest.TestCase):
    def setUp(self):
        data = NPArray([10.0, -103.5, 1000.43, 0.0, 10.0])
        unit = PQ('3s')
        error = NPArray([0.1, 0.2, 4.5, 0.1, 0.2])
        longname = u'Test: FieldContainer H5FileHandler'
        shortname = u'TestH5FC'
        attributes = {'custom1':u'testing1...', 'custom2':u'testing2...'}
        self.fc = FieldContainer(data, unit, error, None, None, longname,
                                 shortname, attributes)
        self.fc.seal()


class FCSaveLoadTestCase(FieldContainerTestCase):
    def setUp(self):
        FieldContainerTestCase.setUp(self)
        osHandle, self.fcFilename = mkstemp(
            suffix = '.h5', prefix = 'pyphantH5FileHandlerTest')
        os.close(osHandle)

    def tearDown(self):
        os.remove(self.fcFilename)

    def testSaveLoad(self):
        handler = H5FH(self.fcFilename, 'w')
        with handler:
            handler.saveDataContainer(self.fc)
            fcLoaded = handler.loadDataContainer(self.fc.id)
        self.assertEqual(self.fc, fcLoaded)


class FCReadOnlyTestCase(FieldContainerTestCase):
    def setUp(self):
        FieldContainerTestCase.setUp(self)
        osHandle, self.rofcFilename = mkstemp(
            suffix = '.h5', prefix = 'pyphantH5FileHandlerTest')
        os.close(osHandle)
        handler = H5FH(self.rofcFilename, 'w')
        with handler:
            handler.saveDataContainer(self.fc)

    def tearDown(self):
        os.remove(self.rofcFilename)

    def testReadOnly(self):
        handler = H5FH(self.rofcFilename, 'r')
        with handler:
            fcLoaded = handler.loadDataContainer(self.fc.id)
        self.assertEqual(self.fc, fcLoaded)


class SampleContainerTestCase(unittest.TestCase):
    def setUp(self):
        data = NPArray([10.0, -103.5, 1000.43, 0.0, 10.0])
        unit = PQ('3s')
        error = NPArray([0.1, 0.2, 4.5, 0.1, 0.2])
        longname = u'Test: FieldContainer H5FileHandler'
        shortname = u'TestH5FC'
        attributes = {'custom1':u'testing1...', 'custom2':u'testing2...'}
        self.fc = FieldContainer(data, unit, error, None, None, longname,
                                 shortname, attributes)
        self.fc.seal()
        fc2 = FieldContainer(NPArray([4003.2, 5.3, 600.9]), PQ('0.2m'), None,
                             None, None, 'FieldContainer 2', 'FC2')
        fc2.seal()
        columns = [self.fc, fc2]
        longname = u'Test: SampleContainer H5FileHandler'
        shortname = u'TestH5SC'
        self.sc = SampleContainer(columns, longname, shortname, attributes)
        self.sc.seal()


class SCSaveLoadTestCase(SampleContainerTestCase):
    def setUp(self):
        SampleContainerTestCase.setUp(self)
        osHandle, self.scFilename = mkstemp(
            suffix = '.h5', prefix = 'pyphantH5FileHandlerTest')
        os.close(osHandle)

    def tearDown(self):
        os.remove(self.scFilename)

    def testSaveLoad(self):
        handler = H5FH(self.scFilename, 'w')
        with handler:
            handler.saveDataContainer(self.sc)
            scLoaded = handler.loadDataContainer(self.sc.id)
        self.assertEqual(self.sc, scLoaded)


class SCReadOnlyTestCase(SampleContainerTestCase):
    def setUp(self):
        SampleContainerTestCase.setUp(self)
        osHandle, self.roscFilename = mkstemp(
            suffix = '.h5', prefix = 'pyphantH5FileHandlerTest')
        os.close(osHandle)
        handler = H5FH(self.roscFilename, 'w')
        with handler:
            handler.saveDataContainer(self.sc)

    def tearDown(self):
        os.remove(self.roscFilename)

    def testReadOnly(self):
        handler = H5FH(self.roscFilename, 'r')
        with handler:
            scLoaded = handler.loadDataContainer(self.sc.id)
        self.assertEqual(self.sc, scLoaded)


class MixedAppendTestCase(SampleContainerTestCase):
    def setUp(self):
        SampleContainerTestCase.setUp(self)
        osHandle, self.appscFilename = mkstemp(
            suffix = '.h5', prefix = 'pyphantH5FileHandlerTest')
        os.close(osHandle)
        handler = H5FH(self.appscFilename, 'w')
        with handler:
            handler.saveDataContainer(self.fc)

    def tearDown(self):
        os.remove(self.appscFilename)

    def testAppend(self):
        handler = H5FH(self.appscFilename, 'a')
        with handler:
            handler.saveDataContainer(self.sc)
            fcLoaded = handler.loadDataContainer(self.fc.id)
            scLoaded = handler.loadDataContainer(self.sc.id)
        self.assertEqual(self.fc, fcLoaded)
        self.assertEqual(self.sc, scLoaded)


class SummaryTestCase(SampleContainerTestCase):
    def setUp(self):
        SampleContainerTestCase.setUp(self)
        osHandle, self.summFilename = mkstemp(
            suffix = '.h5', prefix = 'pyphantH5FileHandlerTest')
        os.close(osHandle)
        handler = H5FH(self.summFilename, 'w')
        with handler:
            handler.saveDataContainer(self.sc)

    def tearDown(self):
        os.remove(self.summFilename)

    def testSummary(self):
        handler = H5FH(self.summFilename, 'r')
        with handler:
            summarydict = handler.loadSummary()
        scsummary = summarydict[self.sc.id]
        fcsummary = summarydict[self.fc.id]
        self.assertEqual(scsummary['id'], self.sc.id)
        self.assertEqual(fcsummary['id'], self.fc.id)
        self.assertEqual(scsummary['longname'], self.sc.longname)
        self.assertEqual(fcsummary['longname'], self.fc.longname)
        self.assertEqual(scsummary['shortname'], self.sc.shortname)
        self.assertEqual(fcsummary['shortname'], self.fc.shortname)
        self.assertEqual(scsummary['creator'], self.sc.creator)
        self.assertEqual(fcsummary['creator'], self.fc.creator)
        self.assertEqual(scsummary['machine'], self.sc.machine)
        self.assertEqual(fcsummary['machine'], self.fc.machine)
        self.assertEqual(scsummary['attributes'], self.sc.attributes)
        self.assertEqual(fcsummary['attributes'], self.fc.attributes)
        self.assertEqual(fcsummary['unit'], self.fc.unit)
        self.assertEqual(scsummary['columns'][0], self.fc.id)
        from pyphant.core.H5FileHandler import im_id
        self.assertEqual(fcsummary['dimensions'], [im_id])


class RecipeTestCase(unittest.TestCase):
    def setUp(self):
        osHandle, self.path = mkstemp(
            suffix='.h5', prefix='pyphantH5RecipeTest')
        os.close(osHandle)

    def tearDown(self):
        os.remove(self.path)

    def testSaveLoadRecipe(self):
        from pyphant.core.CompositeWorker import CompositeWorker
        from pyphant.core.WorkerRegistry import WorkerRegistry
        from itertools import chain
        recipe = CompositeWorker(annotations={'foo':42.0, 'bar':u'Hääh?'})
        wreg = WorkerRegistry.getInstance()
        workerInfos = [t.workerInfos for t in wreg.getToolBoxInfoList()]
        for wInfo in chain(*workerInfos):
            worker = wInfo.createWorker()
            recipe.addWorker(worker)
        with H5FH(self.path, 'w') as handler:
            handler.saveRecipe(recipe)
        with H5FH(self.path, 'r') as handler:
            loadedRecipe = handler.loadRecipe()
        loadedWorkers = loadedRecipe.getWorkers()
        workers = recipe.getWorkers()
        self.assertEqual(len(loadedWorkers), len(workers))
        workerNames = [w.name for w in workers]
        loadedWorkerNames = [w.name for w in loadedWorkers]
        for name in loadedWorkerNames:
            self.assertTrue(name in workerNames)
        self.assertEqual(recipe._annotations, loadedRecipe._annotations)


if __name__ == "__main__":
    import sys
    if len(sys.argv) == 1:
        unittest.main()
    else:
        suite = unittest.TestLoader().loadTestsFromTestCase(
            eval(sys.argv[1:][0]))
        unittest.TextTestRunner().run(suite)



# -*- coding: utf-8 -*-

# Copyright (c) 2006-2008, Rectorate of the University of Freiburg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the Freiburg Materials Research Center,
#   University of Freiburg nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
This module provides an HTTP server (paste) with URL routing (bottle)
"""


import pyphant.core.bottle
from paste import httpserver
from paste.translogger import TransLogger
from threading import Thread
from time import sleep, time
import logging
from urllib2 import (urlopen, URLError, HTTPError)
from mimetypes import MimeTypes


class UnreachableError(Exception):
    pass


class PasteServer(pyphant.core.bottle.ServerAdapter):
    def run(self, handler):
        app = TransLogger(handler, logger=logging.getLogger('pyphant'))
        self.httpserver = httpserver.serve(app, host=self.host,
                                           port=str(self.port),
                                           start_loop=False)
        self.httpserver.serve_forever()


class PasteServerThread(Thread):
    def __init__(self, host, port, app, *args, **kargs):
        Thread.__init__(self, *args, **kargs)
        self.host = host
        self.port = port
        self.app = app

    def run(self):
        self.paste_server = PasteServer(host=self.host, port=self.port)
        try:
            pyphant.core.bottle.run(app=self.app, server=self.paste_server)
        except Exception, exep:
            self.error = exep


class RoutingHTTPServer(object):
    """
    paste server with bottle routing and start/stop/pause methods
    usage:
      server = RoutingHTTPServer()
      server.app : Bottle instance, use it to add routes etc.
      server.url : URL server is listening at
    """

    def __init__(self, host=u'127.0.0.1', port=8080, start=False):
        """
        Sets up the server and (optionally) starts it.
        Arguments:
        - `host`: hostname to listen on
        - `port`: port to listen on
        - `start`: flag that indicates whether to start the server
        """
        self.host = host.lower()
        self.port = int(port)
        self.server_thread = None
        self.mimetypes = MimeTypes()
        self.app = pyphant.core.bottle.Bottle(catchall=False, optimize=True,
                                              autojson=True)
        self.app.serve = False
        self.is_serving = False
        self.logger = logging.getLogger('pyphant')
        if start:
            self.start()

    def _get_url(self):
        return u'http://%s:%d/' % (self.host, self.port)
    url = property(_get_url)

    def start(self):
        """
        Starts or resumes the server.
        """
        if self.is_serving:
            self.app.serve = True
            return
        self.server_thread = PasteServerThread(self.host, self.port, self.app)
        self.server_thread.start()
        self.logger.info("Waiting for server thread to start...")
        while not hasattr(self.server_thread, 'paste_server') \
                  and not hasattr(self.server_thread, 'error'):
            sleep(0.1)
        while not hasattr(self.server_thread.paste_server, 'httpserver') \
                  and not hasattr(self.server_thread, 'error'):
            sleep(0.1)
        isup = False
        t1 = time()
        while not hasattr(self.server_thread, 'error') and not isup \
                  and time() - t1 < 10.0:
            sleep(0.1)
            try:
                stream = None
                try:
                    stream = urlopen(self.url, timeout=3.0)
                except TypeError:
                    stream = urlopen(self.url)
                isup = True
            except HTTPError:
                isup = True
            except (URLError, IOError):
                pass
            finally:
                if stream != None:
                    stream.close()
        if hasattr(self.server_thread, 'error'):
            self.logger.error("Could not start server thread. Reason: %s" \
                                     % self.server_thread.error.__str__())
            error = self.server_thread.error
            self.stop()
            raise error
        if not isup:
            msg = "Server started, but is not responding."
            self.logger.error(msg)
            self.stop()
            raise UnreachableError()
        self.logger.info("Server thread started.")
        self.app.serve = True
        self.is_serving = True

    def pause(self):
        """
        Pauses the server, i.e. denies all requests. Use start() to resume.
        """
        self.app.serve = False

    def stop(self):
        """
        Stops the server.
        """
        self.app.serve = False
        try:
            self.server_thread.paste_server.httpserver.server_close()
        except AttributeError:
            pass
        self.logger.info("Waiting for server thread to stop...")
        self.server_thread.join(10.0)
        if self.server_thread.isAlive():
            self.logger.error("Could not stop server thread!")
        else:
            self.logger.info("Server thread stopped.")
        self.server_thread = None
        self.is_serving = False



# -*- coding: utf-8 -*-

# Copyright (c) 2013, Rectorate of the University of Freiburg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the Freiburg Materials Research Center,
#   University of Freiburg nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

u"""Provides unittest classes for core.Helpers"""

import unittest
from pyphant.core.Helpers import parseFCUnit
from pyphant.quantities import Quantity


class ParseFCUnitTestCase(unittest.TestCase):
    def testFloat(self):
        self.assertEqual(parseFCUnit("2.3e-5"), 2.3e-5)

    def testInt(self):
        self.assertEqual(parseFCUnit("42"), 42.0)

    def testQuantity(self):
        self.assertEqual(parseFCUnit("120 mm"), Quantity("120 mm"))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
from __future__ import with_statement

# Copyright (c) 2006-2009, Rectorate of the University of Freiburg
# Copyright (c) 2009-2010, Andreas W. Liehr (liehr@users.sourceforge.net)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the Freiburg Materials Research Center,
#   University of Freiburg nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

u"""
=================================================================
The **FieldContainer** -- A module storing sampled scalar fields
=================================================================

The *FieldContainer* represents a class of Pyphant's L{DataContainer}
module.

        That is Pyphant's module for *self-desriptive scientific data*
        which is designed to maximise the interoperability of the various
        workers. It can be seen as an interface for exchanging scientific
        information between workers and visualizers. It reproduces the
        self-descriptiveness of the *network Common Data Form* (netCDF). Once
        sealed a DataContainer is immutable but can be identified by
        its *emd5* format which holds information about the origin of
        the container.

The *FieldContainer* stores a sampled scalar Filed.

        It holds an *n-dimensional array* together with
        its *unit* and coordinates of the idependent variable (*dimensions*).
"""


import scipy, copy, hashlib, threading, numpy, StringIO
import os, platform, datetime, socket, urlparse
from pyphant.quantities import (isQuantity, Quantity,_prefixes)
from DataContainer import DataContainer, enc, _logger
from types import NoneType

#Default variables of indices
INDEX_NAMES=[u'i', u'j', u'k', u'l', u'm', u'n']
PREFIXES_METER = copy.deepcopy(_prefixes)
map(lambda r: PREFIXES_METER.remove(r),[('h',  1.e2),('da', 1.e1)])
PREFIXES_METER.append(('',1.0))
PREFIXES = copy.deepcopy(PREFIXES_METER)
map(lambda r: PREFIXES.remove(r),[('d',  1.e-1),('c',  1.e-2)])

class IndexMarker(object):
    hash = hashlib.md5().hexdigest()
    shortname=u"i"
    longname=u"index"
    rawDataBytes = 0
    def seal(self, id=None):
        pass

    def __eq__(self,other):
        if type(self) == type(other):
            return True
        else:
            return False

    def __ne__(self,other):
        return not self.__eq__(other)

    def isValid(self):
        return True

    def getMaskedFC(self, numpymask):
        """
        This method is used to terminate the recursion in
        FieldContainer.getMaskedFC
        """
        return IndexMarker()


#INDEX must be a list of objects providing a .seal(self) method.
#It is used as a marker to terminate the recursive inclusion of FieldContainers.
INDEX=[IndexMarker()]

def generateIndex(i, n,indexNames = INDEX_NAMES):
    u"""Returns a FieldContainer for index variables.
    It stores an index vector (0,...,n-1)^T, whose short name
    will be given by the i. element of list indexNames, if i<=n,
    and 'i_\%i' %i otherwise.
    """
    if i<len(indexNames):
        name=u"%s" % indexNames[i]
    else:
        name=u"i_%i"%i
    return FieldContainer(scipy.arange(0,n), dimensions=INDEX,
                          longname=u"Index", shortname=name)


class DimensionList(list):
    write = True
    def __setitem__(self,i,y):
        if self.write:
            super(DimensionList,self).__setitem__(i,y)
        else:
            raise TypeError, 'Sealed list of dimensions cannot be modified.'

    def __delitem__(self,i):
        if self.write:
            super(DimensionList,self).__delitem__(i,y)
        else:
            raise TypeError, 'Sealed list of dimensions cannot be modified.'

    def seal(self):
        self.write = False

def slice2ind(arg, dim):
    if isinstance(arg, type("")):
        sl = [ "0"+a for a in arg.split(':')] #Hack for quantities. which does not recognize .6 as a number.
        unit = dim.unit
        try:
            hi = Quantity(sl[1])
            try:
                li = Quantity(sl[0])
            except:
                li = Quantity(float(sl[0]),hi.unit)
            try:
                li, hi = [ i.inUnitsOf(unit.unit).value/unit.value for i in [li,hi] ]
            except TypeError:
                raise TypeError("IncompatibleUnits: Dimension %s: %s [%s] can not be sliced with \"%s\"."
                                % (dim.longname, dim.shortname, dim.unit, arg))
        except SyntaxError:
            li, hi = [ float(i)/unit for i in sl ]
        f = dim.data
        intervallElements = numpy.logical_and(f>=li, f<hi)
        if numpy.alltrue(intervallElements):
            start = 0
            end   = len(intervallElements)+1
        else:
            dma = numpy.diff(intervallElements).nonzero()
            if len(dma[0])==0:
                raise NotImplementedError("This slice needs interpolation, which is not implemented yet.")
            if li <= f.min():
                start = 0
                end = dma[0][0]+1
            elif hi > f.max():
                start = dma[0][0]+1
                end = len(intervallElements)+1
            elif hi == f.max():
                start = dma[0][0]+1
                end = len(intervallElements)
            else:
                start, end = dma[0]+1
        return slice(start, end)
    elif isinstance(arg, type(2)):
        return slice(arg, arg+1)
    else:
        return arg

class FieldContainer(DataContainer):
    u"""FieldContainer(data, unit=1, error=None,dimensions=None, longname=u"Sampled Field",
\t\t\t  shortname=u"\\Psi",rescale=False)
\t  Class describing sampled fields:
\t  .data\t\t- Numpy.array representing the sampled field.
\t  .unit\t\t- Quantity object denoting the unit of the sampled field.
\t  .dimensions\t- List of FieldContainer instances
\t\t\t  describing the dimensions of the sampled field.
\t  .data \t- Sampled field stored as numpy.array, which is rescaled to reasonable basic units if option rescale is chosen.
\t  .error\t- Absolut error of the sampled field stored as numpy.array
\t  .longname \t- Notation of the data, e.g. 'electric field',
\t\t\t  which is used for the automatic annotation of charts.
\t  .shortname \t- Symbol of the physical variable in LaTeX notation, e.g. 'E_\\alpha',
\t\t\t  which is also used for the automatic annotation of charts.
\t  .id \t\t- Identifier of Enhanced MD5 (emd5) format
\t\t\t\temd5://NODE/USER/DATETIME/MD5-HASH.TYPESTRING
\t\t\t  which is set by calling method .seal() and
\t\t\t  indicates that the stored information are unchangable.
\t  .label\t- Typical axis description composed from the meta information of the DataContainer.
Concerning the ordering of data matrices and the dimension list consult http://wiki.pyphant.org/xwiki/bin/view/Main/Dimension+Handling+in+Pyphant.
"""
    typeString = u"field"
    def __init__(self, data, unit=1, error=None, mask=None,
                 dimensions=None, longname=u"Sampled Field",
                 shortname=u"\\Psi", attributes=None, rescale=False):
        DataContainer.__init__(self, longname, shortname, attributes)
        self.data = data
        self.mask = mask
        try:
            if isinstance(unit, (str, unicode)):
                unit = unit.replace('^', '**')
            if isinstance(unit, unicode):
                unit = unit.encode('utf-8')
            self.unit = Quantity(unit)
        except:
            try:
                self.unit = Quantity("1"+unit)
            except:
                self.unit = unit
        self.error = error
        if dimensions != None:
            self.dimensions = dimensions
        else:
            N = len(data.shape)-1
            self.dimensions = [generateIndex(N-i,n) for i,n in enumerate(data.shape)]
        if rescale:
            self.rescale()
            for dim in self._dimensions:
                dim.rescale()
        assert self.isValid()

    def _set_dimensions(self,dimensions):
        self._dimensions = DimensionList(dimensions)

    def _get_dimensions(self):
        return self._dimensions
    dimensions = property(_get_dimensions,_set_dimensions)

    def _getLabel(self):
        if len(self._dimensions)>0:
            shortnames = [dim.shortname for dim in self._dimensions]
            shortnames.reverse()
            dependency = '(%s)' % ','.join(shortnames)
        else:
            dependency = ''
        label = u"%s $%s%s$ / %s" % (self.longname.title(), self.shortname, dependency, self.unit)
        try:
            if not isQuantity(self.unit) and self.unit == 1:
                label = u"%s $%s%s$ / a.u." % (self.longname.title(),self.shortname,dependency)
        except:
            pass #just a ScientificPython bug
        return label.replace('1.0 ',r'')#.replace('mu',u'\\textmu{}')
    label=property(_getLabel)

    def _getShortLabel(self):
        if not isQuantity(self.unit) and self.unit == 1:
            if self.longname == 'index':
                label = u"%s $%s$" % (self.longname.title(),self.shortname)
            else:
                label = u"%s $%s$ / a.u." % (self.longname.title(),self.shortname)
        else:
            label =  u"%s $%s$ / %s" % (self.longname.title(), self.shortname, self.unit)
        return label.replace('1.0 ',r'')#.replace('mu',u'\\textmu{}')
    shortlabel=property(_getShortLabel)

    def _getRawDataBytes(self):
        return self.data.nbytes \
               + sum([dim.rawDataBytes for dim in self.dimensions])
    rawDataBytes = property(_getRawDataBytes)

    def __deepcopy__(self, memo):
        self.lock.acquire()
        data=copy.deepcopy(self.data, memo)
        data.setflags(write=True)
        mask=copy.deepcopy(self.mask, memo)
        if mask!=None:
            mask.setflags(write=True)
        error=copy.deepcopy(self.error, memo)
        if error!=None:
            error.setflags(write=True)
        dimensions=copy.deepcopy(self._dimensions, memo)
        res = FieldContainer(data, self.unit, error, mask, dimensions,
                             self.longname, self.shortname)
        self.lock.release()
        return res

    def generateHash(self, m=None):
        if m == None:
            m = hashlib.md5()
        super(FieldContainer, self).generateHash(m)
        #m.update(str(self.data.tolist()))
        m.update(self.data.dumps())
        m.update(str(self.unit))
        if self.error!=None:
            #m.update(str(self.error.tolist()))
            m.update(self.error.dumps())
        if self.mask!=None:
            #m.update(str(self.mask.tolist()))
            m.update(self.mask.dumps())
        [m.update(dim.hash) for dim in self._dimensions]
        return enc(m.hexdigest())

    def seal(self, id=None):
        with self.lock:
            assert self.isValid()
            self.data.setflags(write=False)
            if self.mask!=None:
                self.mask.setflags(write=False)
            if self.error!=None:
                self.error.setflags(write=False)
            if not id:
                self._dimensions.write = False
                for dim in self._dimensions:
                    dim.seal()
            super(FieldContainer, self).seal(id)

    def inUnitsOf(self, other):
        if not isQuantity(self.unit):
            if isQuantity(other.unit):
                raise ValueError("Incompatible Units: self.unit = <%s>, other.unit = <%s>"%(self.unit, other.unit))
            factor = float(self.unit)/float(other.unit)
        elif not isQuantity(other.unit):
            raise ValueError("Incompatible Units: self.unit = <%s>, other.unit = <%s>"%(self.unit, other.unit))
        else:
            if not self.unit.isCompatible(other.unit.unit):
                raise ValueError("Incompatible Units: self.unit = <%s>, other.unit = <%s>"%(self.unit, other.unit))
            factor = self.unit.inUnitsOf(other.unit.unit).value/other.unit.value
        newSelf = copy.deepcopy(self)
        newSelf.data *= factor
        if newSelf.error != None:
            newSelf.error *= factor
        newSelf.unit = copy.deepcopy(other.unit)
        return newSelf

    def rescale(self):
        if isQuantity(self.unit):
            oldUnit = self.unit.inBaseUnits()
        else:
            return
        #Compute decade of field and multiply it to oldUnit
        oldFieldAmplitude = max(abs(numpy.amax(self.data)),abs(numpy.amin(self.data)))
        oldUnit *= oldFieldAmplitude
        #Compute next lower decade
        decade = scipy.log10(oldUnit.value)
        newDecade = 10**(scipy.floor(decade))
        #Find appropriate prefix
        baseUnit=oldUnit.unit.name()
        if baseUnit == 'm':
            prefixes = PREFIXES_METER
        else:
            prefixes = PREFIXES
        prefixCandidates = map(lambda i: (i[0],abs(i[1]-newDecade)),prefixes)
        optPrefix = min([prefix[1] for prefix in prefixCandidates])
        newPrefix = filter(lambda prefix: prefix[1]==optPrefix,prefixCandidates)[0][0]
        newUnitName = newPrefix+baseUnit
        #Convert to new unit
        newUnit = oldUnit.inUnitsOf(newUnitName)
        unitAmplitude = newUnit.value
        if self.data.dtype.name.startswith('int'):
            self.unit = newUnit/oldFieldAmplitude
            return
        self.data *= unitAmplitude/oldFieldAmplitude
        self.unit = newUnit/unitAmplitude

    def __eq__(self, other, rtol=1e-5, atol=1e-8):
        if type(self) != type(other):
            if type(other) != IndexMarker and type(other) != NoneType:
                _logger.debug('Cannot compare objects with different type (%s and %s).' % (type(self),type(other)))
            return False
        if not (self.typeString == other.typeString):
            _logger.debug('The typeString is not identical.')
            return False
        if (self.mask==None) and (other.mask!=None):
            _logger.debug('The mask of the first field container has not been set, while the mask of the second field container is set to %s.' % other.mask)
            return False
        elif  self.mask!=None and (other.mask==None):
            _logger.debug('The mask of the second field container has not been set, while the mask of the first field container is set to %s.' % self.mask)
            return False
        if not (numpy.alltrue(self.mask==other.mask)):
            _logger.debug('The masks are not identical: %s\n%s' % (self.mask,other.mask))
            return False
        if self.mask!=None:
            data = self.data[numpy.logical_not(self.mask)]
            otherData = other.data[numpy.logical_not(other.mask)]
            if self.error!=None:
                error = self.error[numpy.logical_not(self.mask)]
            else:
                error = self.error
            if other.error!=None:
                otherError = other.error[numpy.logical_not(other.mask)]
            else:
                otherError = other.error
        else:
            data = self.data
            error = self.error
            otherData = other.data
            otherError = other.error
        if (isQuantity(self.unit) or isQuantity(other.unit)):
            try:
                if not (self.unit.inBaseUnits().unit == other.unit.inBaseUnits().unit):
                    _logger.debug('The units are different.')
                    return False
            except AttributeError:
                _logger.debug('Cannot compare unit with normed quantity: %s, %s' % (self.unit,other.unit))
                return False
            try:
                scaledData = data*self.unit.value
                scaledOtherData = otherData*other.unit.inUnitsOf(self.unit.unit).value
                if not numpy.allclose(scaledData,scaledOtherData,rtol,atol):
                    if numpy.sometrue(numpy.isnan(scaledData)):
                        _logger.debug('The fields cannot be compared, because some elements of the first field are NaN and the mask has not been set.')
                    if numpy.sometrue(numpy.isnan(scaledOtherData)):
                        _logger.debug('The fields cannot be compared, because some elements of the second field are NaN and the mask has not been set.')
                    else:
                        difference = numpy.abs(scaledData-scaledOtherData)
                        _logger.debug('The scaled fields differ, data-otherData: %s\n%s\n%s' % (difference.max(),
                                                                                       scaledData,
                                                                                       scaledOtherData))
                    return False
            except ValueError:
                _logger.debug('Shape mismatch: %s != %s' % (self.data.shape,other.data.shape))
                return False
            if error!=None:
                scaledError = error*self.unit.value
                if otherError!=None:
                    otherScaledError = otherError*other.unit.inUnitsOf(self.unit.unit).value
                else:
                    _logger.debug('The errors differ: The error of the second argument is none, while the error of the first argument is %s.' % error)
                    return False
                if not numpy.allclose(scaledError,otherScaledError,rtol,atol):
                    _logger.debug('The normed errors differ: %s\n%s' % (scaledError,otherScaledError))
                    return False
        else:
            if not data.dtype.char in ['S','U']:
                try:
                    scaledData = data*self.unit
                    scaledOtherData = otherData*other.unit
                    if not numpy.allclose(scaledData,scaledOtherData,rtol,atol):
                        _logger.debug('The scaled fields differ: %s\n%s'%(scaledData,scaledOtherData))
                        return False
                except ValueError:
                    _logger.debug('Shape mismatch: %s != %s' % (self.data.shape,other.data.shape))
                    return False
                if error==None:
                    if not (otherError==None):
                        _logger.debug('The errors differ: Error of first argument is None, but the error of the second argument is not None.')
                        return False
                else:
                    scaledError = error*self.unit
                    otherScaledError = otherError*other.unit
                    if not numpy.allclose(scaledError,otherScaledError,rtol,atol):
                        _logger.debug('The errors differ: %s\n%s' % (scaledError,otherScaledError))
                        return False
        if not self.attributes == other.attributes:
            _logger.debug('The attribute dictionary differs.')
            return False
        for dimSelf,dimOther in zip(self._dimensions,other.dimensions):
            if dimSelf != dimOther:
                _logger.debug('Different dimensions: %s, %s' % (dimSelf,dimOther))
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __add__(self, other):
        if isinstance(other, FieldContainer):
            if self.error!=None or other.error!=None:
                return NotImplemented
            else:
                error = None
            if len(self._dimensions) != len(other.dimensions):
                return NotImplemented
            for i in xrange(len(self._dimensions)):
                if not self._dimensions[i] == other.dimensions[i]:
                    return NotImplemented
            if isQuantity(self.unit):
                if not isQuantity(other.unit):
                    return NotImplemented
                if not self.unit.isCompatible(other.unit.unit):
                    return NotImplemented
                if self.unit >= other.unit:
                    data = self.data+(other.data*other.unit.value*other.unit.unit.conversionFactorTo(self.unit.unit))/self.unit.value
                    unit = self.unit
                else:
                    data = other.data+(self.data*self.unit.value*self.unit.unit.conversionFactorTo(other.unit.unit))/other.unit.value
                    unit = other.unit
            elif isQuantity(other.unit):
                return NotImplemented
            else:
                data = (self.data*self.unit) + (other.data*other.unit)
                unit = 1.0
            if self.mask==None:
                mask=other.mask
            elif other.mask==None:
                mask=self.mask
            else:
                mask=self.mask+other.mask
            longname = u"Sum of %s and %s." % (self.longname, other.longname)
            shortname = u"%s + %s" % (self.shortname, other.shortname)
            return FieldContainer(data, unit, error, mask,
                                  copy.deepcopy(self._dimensions),
                                  longname, shortname)
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, FieldContainer):
            if self.error!=None or other.error!=None:
                error = other.error + self.error
            else:
                error = None
            if len(self._dimensions) != len(other.dimensions):
                return NotImplemented
            for i in xrange(len(self._dimensions)):
                if not self._dimensions[i] == other.dimensions[i]:
                    return NotImplemented
            if isQuantity(self.unit):
                if not (isQuantity(other.unit) and self.unit.isCompatible(other.unit.unit)):
                    return NotImplemented
                if self.unit >= other.unit:
                    data = self.data - (other.data*other.unit.value*other.unit.unit.conversionFactorTo(self.unit.unit))/self.unit.value
                    unit = self.unit
                else:
                    data = ((self.data*self.unit.value*self.unit.unit.conversionFactorTo(other.unit.unit))/other.unit.value) - other.data
                    unit = other.unit
            else:
                if isQuantity(other.unit):
                    return NotImplemented
                data = (self.data*self.unit) - (other.data*other.unit)
                unit = 1.0
            if self.mask==None:
                mask=other.mask
            elif other.mask==None:
                mask=self.mask
            else:
                mask=self.mask+other.mask
            longname = u"Difference of %s and %s." % (self.longname, other.longname)
            shortname = u"%s - %s" % (self.shortname, other.shortname)
            return FieldContainer(data, unit, error, mask,
                                  copy.deepcopy(self._dimensions),
                                  longname, shortname)
        return NotImplemented

    def __str__(self):
        deps = [ dim for dim in self._dimensions if type(dim)!=type(IndexMarker()) ]
        report  = "\nFieldContainer %s of shape %s with field\n%s\n"% (self.label,self.data.shape,self.data)
        if self.error != None:
            report += ", error\n%s\n" % self.error
        if self.mask !=None:
            report += ", mask\n%s\n" % self.mask
        if len(deps)>0:
            report += 'depending on dimensions %s\n' % deps
        if len(self.attributes)>0:
            report += 'and attributes\n%s\n' % self.attributes
        return report

    def __repr__(self):
        return self.__str__()

    def __getitem__(self, args):
        if isinstance(args, type("")):
            args=[args]
        if isinstance(args, type(1)):
            if args>=len(self.data):
                raise IndexError, 'index out of bound'
        try:
            len(args)
        except:
            args = [args]
        args = [ slice2ind(arg, self._dimensions[dim]) for dim, arg in enumerate(args) ]
        data = self.data[args]
        attributes = copy.deepcopy(self.attributes)
        mask = None
        error = None
        dimensions = []
        for i,l in enumerate(data.shape[:len(args)]):
            dim = self._dimensions[i]
            if l==1:
                attributes[dim.longname] = (dim.shortname, dim.data[args[i]].squeeze()*dim.unit)
            else:
                if isinstance(dim, IndexMarker):
                    dimensions.append(dim)
                else:
                    dimensions.append(dim[args[i],])
        for i in xrange(len(args),len(data.shape)):
            dimensions.append(copy.deepcopy(self._dimensions[i]))
        if data.shape != (1,):
            data = data.squeeze()
            if self.mask!=None:
                mask = self.mask[args].squeeze()
            if self.error!=None:
                error = self.error[args].squeeze()
        else:
            if self.mask!=None:
                mask = self.mask[args]
            if self.error!=None:
                error = self.error[args]
        field = FieldContainer(data, dimensions=dimensions,
                               longname=self.longname,
                               shortname=self.shortname,
                               mask=mask,
                               error=error,
                               unit=self.unit,
                               attributes=attributes)
        return field

    def isValid(self):
        # Valid dimensions?
        if (not (len(self._dimensions)==1
                 and isinstance(self._dimensions[0], IndexMarker)) #IndexMarkers are valid and...
            and not (self.data.shape == (1,) and len(self._dimensions)==0)): #...so are zero dim fields.
            dimshape = []
            for d in self._dimensions:
                if len(d.data.shape)>1:
                    _logger.debug("Dimension %s is not 1-d." % d.longname)
                    return False
                dimshape.append(d.data.shape[0])
            if self.data.shape!=tuple(dimshape):
                _logger.debug("Shape of data %s and of dimensions %s do not match for field\n:%s" %
                              (self.data.shape, dimshape, self))
                return False
            for d in self._dimensions:
                if not d.isValid():
                    _logger.debug("Invalid dimension %s."%d.longname)
                    return False
        # Valid mask?
        if (self.mask!=None) and (self.data.shape!=self.mask.shape):
            _logger.debug("Shape of data %s and of mask %s do not match."%(self.data.shape, self.mask.shape))
            return False
        # Valid error?
        if (self.error!=None) and (self.data.shape!=self.error.shape):
            _logger.debug("Shape of data %s and of error %s do not match."%(self.data.shape, self.error.shape))
            return False
        return True

    def isIndex(self):
        return self.dimensions==INDEX

    def isIndependent(self):
        for d in self.dimensions:
            if not d.isIndex():
                return False
        return True

    def isLinearlyDiscretised(self):
        for dim in self.dimensions:
            d = numpy.diff(dim.data)
            if not numpy.alltrue(numpy.logical_not(d-d[0])):
                return False
        return True

    def getMaskedFC(self, numpymask):
        """
        Return an unsealed FieldContainer instance that emerges from
        this instance by clipping along the primary axis according to the
        entries in the parameter numpymask, i.e. all entries where numpymask
        is set to False are discarded.
        numpymask -- Numpy array with Boolean values
        """
        mdims = []
        for dim in self.dimensions:
            if mdims == []:
                # only primary axis has to be masked:
                mdims.append(dim.getMaskedFC(numpymask))
            else:
                mdims.append(copy.deepcopy(dim))
        to_mask = [self.data, self.error, self.mask]
        masked = []
        for item in to_mask:
            if item != None:
                masked.append(item[numpymask])
            else:
                masked.append(None)
        return FieldContainer(masked[0],
                              copy.deepcopy(self.unit),
                              masked[1],
                              masked[2],
                              mdims,
                              longname=self.longname,
                              shortname=self.shortname,
                              attributes=copy.deepcopy(self.attributes),
                              rescale=False)

    maskedData = property( lambda self: numpy.ma.array(self.data, mask=self.mask) )
    maskedError = property( lambda self: numpy.ma.array(self.error, mask=self.mask) )




# -*- coding: utf-8 -*-

# Copyright (c) 2006-2007, Rectorate of the University of Freiburg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the Freiburg Materials Research Center,
#   University of Freiburg nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

u"""
The Composite worker acts as a container for graphs of workers.
At the moment these can only be entire recipes, but it is planned
that they can be subgraphs, too.
"""


import EventDispatcher, Worker, Connectors, Param
import copy
import pkg_resources


class CompositeWorkerChangedEvent(object):
    def __init__(self, worker, message, data=None):
        self.worker = worker
        self.message = message
        self.data = data

class WorkerAddedEvent(CompositeWorkerChangedEvent):
    def __init__(self, worker, data=None):
        CompositeWorkerChangedEvent.__init__(self, worker,
                                             "Worker Added", data)

class WorkerRemovedEvent(CompositeWorkerChangedEvent):
    def __init__(self, worker, data=None):
        CompositeWorkerChangedEvent.__init__(self, worker,
                                             "Worker Removed", data)

class ConnectionEvent(CompositeWorkerChangedEvent):
    def __init__(self, plug, socket, message, data=None):
        CompositeWorkerChangedEvent.__init__(self, socket.worker,
                                             message, data)
        self.plug = plug
        self.socket = socket
class ConnectionCreatedEvent(ConnectionEvent):
    def __init__(self, plug, socket, data=None):
        ConnectionEvent.__init__(self, plug, socket,
                                 "Connection created", data)
class ConnectionDestroyedEvent(ConnectionEvent):
    def __init__(self, plug, socket, data=None):
        ConnectionEvent.__init__(self, plug, socket,
                                 "Connection destroyed", data)
class ConnectorsExternalizationStateChangedEvent(CompositeWorkerChangedEvent):
    def __init__(self, worker, connector, data=None):
        CompositeWorkerChangedEvent.__init__(
            self, worker, "Connectors externalization state changed", data
            )
        self.connector = connector

class CompositeWorker(Worker.Worker):
    API = 2
    VERSION = 1
    REVISION = pkg_resources.get_distribution("pyphant").version
    name = "Composite"
    _params = [("noSockets", "Number of sockets", 0, None),
               ("noPlugs", "Number of plugs", 0, None)]

    def _id(self):
        if self.parent == None:
            return ""
        else:
            return super(CompositeWorker, self)._id()
    id = property(_id)

    def inithook(self):
        self._workers = []
        self._sources = []
        self._sinks = []
        self._eventDispatcher = EventDispatcher.EventDispatcher()
        for i in xrange(self.paramNoSockets.value):
            name = "socket%i" % i
            socketProxy = Connectors.ConnectorProxy(self, True, name)
            setattr(self, 'socket'+self.upperFirstLetter(name), socketProxy)
            self._sockets[name] = socketProxy
        for i in xrange(self.paramNoPlugs.value):
            name = "plug%i" % i
            plugProxy = Connectors.ConnectorProxy(self, False, name)
            setattr(self, 'plug'+self.upperFirstLetter(name), plugProxy)
            self._plugs[name] = plugProxy

    def addWorker(self, worker, data=None):
        if not self.checkWorkerName(worker, worker.getParam('name').value):
            raise ValueError(u'Duplicate worker name')
        worker.registerParamListener(self.vetoWorkerName,
                                     'name', Param.ParamChangeRequested)
        self._workers.append(worker)
        if len(worker.getPlugs()) == 0:
            self._sinks.append(worker)
        if len(worker.getSockets()) == 0:
            self._sources.append(worker)
        self._notifyListeners(WorkerAddedEvent(worker, data))

    def removeWorker(self, worker, data=None):
        self._workers.remove(worker)
        if worker in self._sources:
            self._sources.remove(worker)
        if worker in self._sinks:
            self._sinks.remove(worker)
        self._notifyListeners(WorkerRemovedEvent(worker, data))

    def getWorker(self, name):
        cands = [w for w in self._workers if w.getParam('name').value == name]
        if len(cands) > 1:
            raise RuntimeError(
                "Ambiguous worker name <%s> in CompositeWorker <%s>."
                % (name, self)
                )
        if len(cands) == 0:
            return None
        return cands[0]

    def getWorkers(self, desiredWorker=''):
        if desiredWorker == '':
            result = self._workers
        else:
            result = [w for w in self._workers
                      if w.__class__.__name__ == desiredWorker]
            if result == []:
                raise ValueError("Recipe does not contain Worker %s"
                                 % desiredWorker)
        return result

    def findConnectorForId(self, id):
        splittedId = id.split('.', 1)
        if len(splittedId) == 1:
            return super(CompositeWorker, self).findConnectorForId(
                splittedId[0]
                )
        else:
            w = self.getWorker(splittedId[0])
            return w.findConnectorForId(splittedId[1])

    def getAllPlugs(self):
        return sum([w.getPlugs() for w in self.getWorkers()], [])

    def getOpenSocketsForPlug(self, plug):
        walker = self.createCompositeWorkerWalker()
        return sum(walker.visit(lambda w:
                                [s for s in w.getSockets()
                                 if not s.isFull()],
                                [plug.worker]),
                   [])

    def getSources(self):
        return self._sources

    def getSinks(self):
        return self._sinks

    #pickle
    def __getstate__(self):
        pdict = copy.copy(self.__dict__)
        pdict['_eventDispatcher'] = EventDispatcher.EventDispatcher()
        return pdict

    def generateEvents(self, listenerDict):
        map(lambda worker: listenerDict[WorkerAddedEvent](
                WorkerAddedEvent(worker)
                ), self._workers)
        def connectionInformer(worker):
            for socket in worker.getSockets():
                if socket.isFull():
                    [ (issubclass(ConnectionCreatedEvent, x)
                       and l(ConnectionCreatedEvent(socket.getPlug(), socket)))
                      for (x, l) in listenerDict.items() ]
        connectionInformer(self)
        map(connectionInformer, self._workers)

    def createCompositeWorkerWalker(self):
        return CompositeWorkerWalker(self)

    def registerListener(self, listener,
                         eventType=CompositeWorkerChangedEvent):
        self._eventDispatcher.registerListener( listener, eventType)

    def unregisterListener(self, listener,
                           eventType=CompositeWorkerChangedEvent):
        self._eventDispatcher.unregisterListener( listener, eventType )

    def _notifyListeners(self, event):
        self._eventDispatcher.dispatchEvent(event)

    def connectionCreated(self, plug, socket):
        if self.parent:
            self.parent.connectionCreated(plug, socket)
        self._notifyListeners(ConnectionCreatedEvent(plug, socket))

    def connectionDestroyed(self, plug, socket):
        if self.parent:
            self.parent.connectionDestroyed(plug, socket)
        self._notifyListeners(ConnectionDestroyedEvent(plug, socket))

    def workersConnectorStateChanged(self, worker, connector):
        self._notifyListeners(
            ConnectorsExternalizationStateChangedEvent(worker, connector)
            )

    def vetoWorkerName(self, paramChangeEvent):
        if not self.checkWorkerName(paramChangeEvent.param.worker,
                                    paramChangeEvent.newValue):
            raise Param.VetoParamChange(paramChangeEvent)

    def checkWorkerName(self, worker, name):
        ws = [w for w in self._workers
              if w.getParam('name').value == name]
        if ws == [] or ws[0] == worker:
            return True
        else:
            return False

class CompositeWorkerWalker(object):
    def __init__(self, compositeWorker):
        self._compositeWorker = compositeWorker

    def visit(self, visitor, start):
        visited = []
        toVisit = start
        while toVisit:
            worker = toVisit.pop(0)
            yield visitor(worker)
            visited.append(worker)
            for socket in worker.getSockets():
                plug = socket.getPlug()
                if plug:
                    worker = plug.worker
                    if worker not in visited:
                        toVisit.append(worker)





# -*- coding: utf-8 -*-

# Copyright (c) 2006-2010, Rectorate of the University of Freiburg
# Copyright (c) 2010, Andreas W. Liehr (liehr@users.sourceforge.net)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the Freiburg Materials Research Center,
#   University of Freiburg nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import with_statement

u"""Provides unittest classes for SQLiteWrapper.
"""


import unittest
import pyphant.core.SQLiteWrapper
from pyphant.quantities import Quantity
from pyphant.core.H5FileHandler import (im_id, im_summary)

class TestQuantity2powers(unittest.TestCase):
    def testPowersOfUnits(self):
        expected = (1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0)
        result = pyphant.core.SQLiteWrapper.quantity2powers(Quantity('1m'))
        self.assertEqual(expected,result,"The number of basic units determining a quantity does not fit. The SQTLiteWrapper is expecting")
        
    def testChangedBaseUnits(self):
        """Test if function quantity2powers checks the expected number of unit powers as given by Quantity.unit.powers
        in order to avoid clashes with the SQLite tables."""
        class Unit:
            powers = (1, 0, 0)
        class fakedQuantity:
            """Faked Quantity class of TestQuantity2powers unittest."""
            def __init__(self):
                self.unit = Unit()
                self.__class__ = Quantity
        quantity = fakedQuantity()
        self.assertRaises(AssertionError,pyphant.core.SQLiteWrapper.quantity2powers,quantity)

class SQLiteWrapperTestCase(unittest.TestCase):
    def setUp(self):
        import tempfile
        import os
        self.dir = tempfile.mkdtemp()
        self.dbase = self.dir + os.path.sep + "testcase.sqlite3"
        self.wrapper = pyphant.core.SQLiteWrapper.SQLiteWrapper(self.dbase)
        with self.wrapper:
            self.wrapper.setup_dbase()
        self.summary = {'id':'emd5://PC/aheld/'
                        '2009-01-01_12:00:00.123456/12345678910.field',
                        'longname':'name',
                        'shortname':'sn', 'creator':'aheld',
                        'date':'2009-01-01_12:00:00.123456',
                        'machine':'PC', 'type':'field',
                        'unit':Quantity('10.03e-8 mm**-1'),
                        'attributes':{'attribute1':'bla1',
                                      'attribute2':'bla2'},
                        'hash':'12345678910',
                        'dimensions':[im_id]}
        self.sc_summary = self.summary.copy()
        self.sc_summary['id'] = self.summary['id'][:-5] + 'sample'
        self.sc_summary.pop('unit')
        self.sc_summary.pop('dimensions')
        self.sc_summary['columns'] = [self.summary['id'], self.summary['id']]
        self.sc_summary['type'] = 'sample'
        self.sc_summary['longname'] = u'name2'
        self.sc_summary['shortname'] = u'sn2'

    def tearDown(self):
        import os
        os.remove(self.dbase)
        os.removedirs(self.dir)

    def testAll(self):
        assert pyphant.core.SQLiteWrapper.date2dbase('2009') \
            == '2009-01-01_00:00:00.000000'
        with self.wrapper:
            id = self.summary['id']
            assert not self.wrapper.has_entry(id)
            self.wrapper.set_entry(im_summary, None)
            assert self.wrapper.has_entry(im_summary['id'])
            self.wrapper.set_entry(self.summary, 'storage1')
            assert self.wrapper.has_entry(id)
            rowwrapper = self.wrapper[id]
            for key, value in self.summary.iteritems():
                assert value == rowwrapper[key]
            assert rowwrapper['storage'] == 'storage1'
            rowwrapper['storage'] = 'storage2'
            assert rowwrapper['storage'] == 'storage2'
            try:
                bad = rowwrapper['bad']
                assert False
            except ValueError:
                pass
            self.wrapper.set_entry(self.summary, 'storage3')
            assert rowwrapper['storage'] == 'storage2'
            self.wrapper.set_entry(self.sc_summary, 'storage4')
            rowwrapper = self.wrapper[self.sc_summary['id']]
            assert rowwrapper['columns'] == self.sc_summary['columns']
            emd5list = self.wrapper.get_emd5_list()
            assert len(emd5list) == 3
            assert self.summary['id'] in emd5list
            assert self.sc_summary['id'] in emd5list
            keys = self.wrapper.common_result_keys
            dictionary = self.summary.copy()
            dictionary.pop('date')
            dictionary.pop('dimensions')
            dictionary.pop('unit')
            dictionary.pop('type')
            dictionary.pop('attributes')
            dictionary['storage'] = 'storage2'
            dictionary['date_from'] = u'2009-01-01_12:00:00.000000'
            dictionary['date_to'] = u'2009-01-01_12:00:00.200000'
            search_result = self.wrapper.get_andsearch_result(
                keys, dictionary, order_by='type', limit=10, offset=0)
            assert len(search_result) == 1
            expected = [(u'name', u'sn', u'PC', u'aheld', u'12345678910',
                         u'2009-01-01_12:00:00.123456', u'emd5://PC/aheld/'\
                             u'2009-01-01_12:00:00.123456/12345678910.field',
                         u'field', u'storage2')]
            assert search_result == expected
            dictionary['type'] = 'field'
            search_result = self.wrapper.get_andsearch_result(keys, dictionary)
            assert search_result == expected
            search_result = self.wrapper.get_andsearch_result(
                ['longname'], order_by='longname')
            assert search_result == [(u'index', ), (u'name', ), (u'name2', )]
            search_result = self.wrapper.get_andsearch_result(
                ['longname'], {'unit':Quantity(20, '1/m'),
                               'type':'field'})
            assert search_result == [(u'name', )]
            search_result = self.wrapper.get_andsearch_result(
                ['unit'], {'type':'field', 'longname':'name'})
            assert search_result == [(Quantity('10.03e-8 mm**-1'), )]
            search_result = self.wrapper.get_andsearch_result(
                ['latex_unit'], {'type':'field', 'longname':'name'})
            #assert search_result == [(..., )]
            search_result = self.wrapper.get_andsearch_result(
                ['longname'], {'attributes':{'attribute2':'bla2'}},
                order_by='longname')
            assert search_result == [(u'name', ), (u'name2', )]
            search_result = self.wrapper.get_andsearch_result(
                ['longname'], {'attributes':{'attribute2':'bla2',
                                             'attribute1':'bla1'}},
                order_by='longname')
            assert search_result == [(u'name', ), (u'name2', )]
            any_value = self.wrapper.any_value
            search_result = self.wrapper.get_andsearch_result(
                ['longname'], {'attributes':{'attribute1':any_value,
                                             'attribute2':'bla2'}})
            assert search_result == [(u'name', ), (u'name2', )]
            search_result = self.wrapper.get_andsearch_result(
                ['longname'], {'type':'sample',
                               'columns':[{'unit':Quantity(2.0, '1/m')}
                                          , {}]})
            assert search_result == [(u'name2', )]
            search_result = self.wrapper.get_andsearch_result(
                ['longname'],
                {'type':'field',
                 'dimensions':[{'longname':im_summary['longname']}],
                 'longname':self.summary['longname']})
            assert search_result == [(u'name', )]
            rowwrapper = self.wrapper[id]
            assert rowwrapper['dimensions'] == self.summary['dimensions']
            search_result = self.wrapper.get_andsearch_result(
                ['longname'], {'type':'field', 'dim_of':{'id':id}})
            assert search_result == [(im_summary['longname'], )]
            search_result = self.wrapper.get_andsearch_result(
                ['shortname'], {'type':'field',
                                'col_of':{'id':self.sc_summary['id']}})
            assert search_result == [(self.summary['shortname'], )]


if __name__ == "__main__":
    import sys
    if len(sys.argv) == 1:
        unittest.main()
    else:
        suite = unittest.TestLoader().loadTestsFromTestCase(
            eval(sys.argv[1:][0]))
        unittest.TextTestRunner().run(suite)



# -*- coding: utf-8 -*-

# Copyright (c) 2006-2007, Rectorate of the University of Freiburg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the Freiburg Materials Research Center,
#   University of Freiburg nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

u"""
"""


import OneLineStringField
import BoundedIntegerTextField
import CheckBox
import ListSelect
import FileButton
from pyphant.core import Connectors, Param
import wx


class NullVal(wx.PyValidator):
    def Clone(self):
        return NullVal()

    def Validate(self, win):
        return True


class ParamValidator(wx.PyValidator):
    def __init__(self, param):
        wx.PyValidator.__init__(self)
        self.param = param

    def Clone(self):
        return ParamValidator(self.param)

    def Validate(self, win):
        ctrl = self.GetWindow()
        newValue = ctrl.getValue()
        try:
            self.param.value = newValue
        except Param.VetoParamChange, e:
            wx.MessageBox("%s" % str(e),
                          "Invalid Parameter %s" % self.param.name)
            ctrl.SetFocus()
            ctrl.Refresh()
            return False
        return True

    def TransferToWindow(self):
        ctrl = self.GetWindow()
        ctrl.SetValue(self.param.value)
        True

    def TransferFromWindow(self):
        ctrl = self.GetWindow()
        self.param.value = ctrl.getValue()
        True


class ParamVisReg:
    def __init__(self):
        self._visualizers={}
        register = [
            (0, None, BoundedIntegerTextField.BoundedIntegerTextField),
            (" ", None, OneLineStringField.OLSF),
            (" ", Connectors.SUBTYPE_FILE, FileButton.FileButton),
            (True, None, CheckBox.CheckBox),
            ([], None, ListSelect.ListSelect),
            ]
        for prototype, subtype, visualizer in register:
            self.setVisualizer(type(prototype), subtype, visualizer)

    def setVisualizer(self, type, subtype, visualizer):
        try:
            typedict=self._visualizers[type]
        except KeyError:
            typedict={}
            self._visualizers[type]=typedict
        typedict[subtype]=visualizer

    def createVisualizerFor(self, parent, param, validator=None):
        if validator==None:
            validator=ParamValidator(param)
        return self._visualizers[param.valueType][param.subtype](
            parent, param, validator=validator)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from setuptools.command.test import test

VERSION = '1.0b3'


class EnsureAggUnittests(test):
    def run_tests(self):
        from pyphant.mplbackend import ensure_mpl_backend
        ensure_mpl_backend('agg')
        return test.run_tests(self)


setup(
    name='pyphant',
    version=VERSION,
    description='Workflow modelling app',
    author='Alexander Held, Andreas W. Liehr, Klaus Zimmermann',
    author_email='alexander.held@fmf.uni-freiburg.de',
    maintainer='Alexander Held',
    maintainer_email='alexander.held@fmf.uni-freiburg.de',
    license="BSD",
    url='http://pyphant.sourceforge.net/',
    install_requires=[
        'sogl>=0.2.1',
        'paste',
        'simplejson',
        'matplotlib',
        'numpy',
        'scipy',
        'tables',
        'wxPython',
        'egenix-mx-base',
        'configobj',
        ],
    packages=find_packages(),
    entry_points={'gui_scripts': [
        'wxPyphant = pyphant.wxgui2.wxPyphantApplication:startWxPyphant'
        ]},
    include_package_data=True,
    cmdclass={'test': EnsureAggUnittests},
    test_suite='pyphant.tests'
    )



# -*- coding: utf-8 -*-

# Copyright (c) 2006-2009, Rectorate of the University of Freiburg
# Copyright (c) 2009, Andreas W. Liehr (liehr@users.sourceforge.net)
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the Freiburg Materials Research Center,
#   University of Freiburg nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

u"""
This module provides the facilities to persist a Recipe including any
generated data into a PyTables structure or a HDF5 file.
A rudimentary description of the used file-format follows:
.:Attribut
+:group
#:table
*:array

/results/resId
.longname
.shortname
[field|sample]

sample:
.desc = sample.desc [pickled scipy.dtype]
#data(descDict["col_"+name.encode("utf-8")]=Col(dtype=str(desc.fields[name][0]).capitalize()))
#units(unit,name)

field:
*data
*error
.unit = repr(field.unit)
UNLESS field.dimensions==INDEX:
#dimensions(hash, id)
"""


import tables, datetime
import sys
from pyphant.core import (CompositeWorker, DataContainer)
from tables import StringCol, Col
from pyphant.quantities import Quantity
from pyphant.quantities import Quantity as PhysicalQuantity
import scipy
import logging
_logger = logging.getLogger("pyphant")

_reservedAttributes = ('longname', 'shortname', 'columns', 'creator', 'machine')

class Connection(tables.IsDescription):
    destinationWorker = tables.StringCol(len("worker_"+str(sys.maxint))+1)
    destinationSocket = tables.StringCol(64)

##########################################################################
# Saving Part
##########################################################################
def saveExecutionOrder(h5, order):
    if 'executionOrder' in h5.root:
        executionOrderGroup = h5.root.executionOrder
    else:
        executionOrderGroup = h5.createGroup('/', 'executionOrder')
    import md5
    m = md5.new()
    m.update(str(sorted(order[0].items()))) # Call 'sorted' to
                                            # normalize representation
                                            # of order[0]
    m.update(str(sorted(order[1])))
    name = "pre_"+m.hexdigest()
    orderGroup = h5.createGroup(executionOrderGroup, name)
    class InputDescription(tables.IsDescription):
        socket = StringCol(max(map(len, order[0].keys())))
        data = StringCol(max(map(len, order[0].values())))
    input = h5.createTable(orderGroup, 'input', InputDescription, "Socket Map")
    m = input.row
    for s, d in order[0].iteritems():
        m['socket'] = s
        m['data'] = d
        m.append()
    input.flush()
    orderGroup._v_attrs.resultPlug = order[1]

def saveRecipe(h5, recipe, saveResults=True):
    recipeGroup = h5.createGroup("/", "recipe")
    h5.createGroup("/", "results")
    h5.setNodeAttr(recipeGroup, "Annotations", recipe._annotations)
    workers = recipe.getWorkers()
    for worker in workers:
        saveWorker(h5, recipeGroup, worker, saveResults)

def saveWorker(h5, recipeGroup, worker, saveResults=True):
    workerGroup = h5.createGroup(recipeGroup, "worker_"+str(hash(worker)))
    saveBaseAttributes(h5, workerGroup, worker)
    savePlugs(h5, workerGroup, worker, saveResults)
    saveParameters(h5, workerGroup, worker)

def saveParameters(h5, workerGroup, worker):
    paramGroup = h5.createGroup(workerGroup, "parameters")
    for (paramName, param) in worker._params.iteritems():
        h5.setNodeAttr(paramGroup, paramName, param.value)

def savePlugs(h5, workerGroup, worker, saveResults=True):
    plugs = h5.createGroup(workerGroup, "plugs")
    for (plugName, plug) in worker._plugs.iteritems():
        plugGroup = h5.createGroup(plugs, plugName)
        if plug.resultIsAvailable() and saveResults:
            resId = saveResult(plug._result, h5)
            h5.setNodeAttr(plugGroup, "result", resId)
        connectionTable = h5.createTable(plugGroup, 'connections', Connection, expectedrows=len(plug._sockets))
        connection=connectionTable.row
        for socket in plug._sockets:
            connection['destinationWorker'] = "worker_"+str(hash(socket.worker))
            connection['destinationSocket'] = socket.name
            connection.append()
        connectionTable.flush()

def saveBaseAttributes(h5, workerGroup, worker):
    h5.setNodeAttr(workerGroup, "module", worker.__class__.__module__)
    h5.setNodeAttr(workerGroup, "clazz", worker.__class__.__name__)
    h5.setNodeAttr(workerGroup, "WorkerAPIVersion", worker.API)
    h5.setNodeAttr(workerGroup, "WorkerVersion", worker.VERSION)
    h5.setNodeAttr(workerGroup, "WorkerRevision", worker.REVISION)
    h5.setNodeAttr(workerGroup, "Annotations", worker._annotations)

def saveResult(result, h5):
    hash, uriType = DataContainer.parseId(result.id)
    resId = u"result_"+hash
    try:
        resultGroup = h5.getNode("/results/"+resId)
    except tables.NoSuchNodeError, e:
        resultGroup = h5.createGroup("/results", resId, result.id.encode("utf-8"))
        if uriType=='field':
            saveField(h5, resultGroup, result)
        elif uriType=='sample':
            saveSample(h5, resultGroup, result)
        else:
            raise KeyError, "Unknown UriType %s in saving result %s." % (uriType, result.id)
    return resId


def saveSample(h5, resultGroup, result):
    h5.setNodeAttr(resultGroup, "longname", result.longname.encode("utf-8"))
    h5.setNodeAttr(resultGroup, "shortname", result.shortname.encode("utf-8"))
    h5.setNodeAttr(resultGroup, "creator", result.creator.encode("utf-8"))
    h5.setNodeAttr(resultGroup, "machine", result.machine.encode("utf-8"))
    for key,value in result.attributes.iteritems():
        if key in _reservedAttributes:
            raise ValueError("Attributes should not be named %s, "
                             "but one was in fact called %s!"
                             % (str(_reservedAttributes), key))
        h5.setNodeAttr(resultGroup,key,value)
    #Store fields of sample Container and gather list of field IDs
    columns = []
    for column in result.columns:
        columns.append(saveResult(column,h5))
    h5.setNodeAttr(resultGroup, "columns", columns)

def saveField(h5, resultGroup, result):
    def dump(inputList):
        def conversion(arg):
            if type(arg) == type(u' '):
                return arg.encode('utf-8')
            else:
                return arg.__repr__()
        if type(inputList)==type([]):
            return map(conversion,inputList)
        else:
            return map(dump,inputList)
    if result.data.dtype.char in ['U','O']:
        unicodeData = scipy.array(dump(result.data.tolist()))
        h5.createArray(resultGroup, "data", unicodeData, result.longname.encode("utf-8"))
    else:
        h5.createArray(resultGroup, "data", result.data, result.longname.encode("utf-8"))
    for key,value in result.attributes.iteritems():
        h5.setNodeAttr(resultGroup.data,key,value)
    h5.setNodeAttr(resultGroup, "longname", result.longname.encode("utf-8"))
    h5.setNodeAttr(resultGroup, "shortname", result.shortname.encode("utf-8"))
    h5.setNodeAttr(resultGroup, "creator", result.creator.encode("utf-8"))
    h5.setNodeAttr(resultGroup, "machine", result.machine.encode("utf-8"))

    if result.error != None:
        h5.createArray(resultGroup, "error", result.error,
                       (u"Error of "+result.longname).encode("utf-8"))
    if result.mask != None:
        h5.createArray(resultGroup, "mask", result.mask,
                       (u"Mask of "+result.longname).encode("utf-8"))
    h5.setNodeAttr(resultGroup, "unit", repr(result.unit).encode("utf-8"))
    if result.dimensions!=DataContainer.INDEX:
        idLen=max([len(dim.id.encode("utf-8")) for dim in result.dimensions])
        dimTable = h5.createTable(resultGroup, "dimensions",
                                  {"hash":StringCol(32), "id":StringCol(idLen)},
                                  (u"Dimensions of "+result.longname).encode("utf-8"),
                                  expectedrows=len(result.dimensions))
        for dim in result.dimensions:
            d = dimTable.row
            d["hash"]=dim.hash.encode("utf-8")
            d["id"]=dim.id.encode("utf-8")
            d.append()
            saveResult(dim, h5)
        dimTable.flush()

##########################################################################
# Loading Part
##########################################################################
def instantiateWorker( parent, workerGroup ):
    annotations = workerGroup._v_attrs.Annotations
    module = workerGroup._v_attrs.module
    exec "import "+module
    worker = eval(module+"."+workerGroup._v_attrs.clazz+"(parent, annotations)")
    return worker

def restoreParamsToWorker(worker, workerGroup):
    try:
        worker.refreshParams()
    except:
        _logger.warning(
            u"Attempted refreshParam failed for %s. Check Parameters!" % (
                worker.name,
                ),
            exc_info=True
            )
    for paramName in workerGroup.parameters._v_attrs._v_attrnamesuser:
        param = getattr(workerGroup.parameters._v_attrs, paramName)
        if type(param) == scipy.ndarray:
            param = unicode(param)
        elif type(param) == scipy.string_:
            param = str(param)
        elif type(param) == scipy.int32:
            param = int(param)
        try:
            worker.getParam(paramName).overrideValue(param)
        except KeyError:
            _logger.warning(
                u'Could not restore "%s" to parameter: "%s"' % (
                    param, paramName
                    )
                )

def restoreParamsToWorkerParentsFirst(worker, worker2group, visited):
    if worker in visited:
        return
    parents = [s.getPlug().worker for s in worker.getSockets() if s.isFull()]
    for parent in parents:
        restoreParamsToWorkerParentsFirst(parent, worker2group, visited)
    restoreParamsToWorker(worker, worker2group[worker])
    visited.add(worker)

def restoreParamsToWorkers(recipeGroup, workers):
    visited = set()
    worker2group = dict([(workers[wg._v_name], wg) for wg in recipeGroup])
    for worker in worker2group:
        restoreParamsToWorkerParentsFirst(worker, worker2group, visited)

def loadRecipe(h5):
    recipeGroup = h5.root.recipe
    try:
        annotations = recipeGroup._v_attrs.Annotations
    except AttributeError:
        annotations = {}
    recipe = CompositeWorker.CompositeWorker(annotations=annotations)
    workers = {}
    createWorkerGraph(recipeGroup, workers, recipe)
    restoreResultsToWorkers(recipeGroup, workers, h5)
    restoreParamsToWorkers(recipeGroup, workers)
    return recipe

def createWorkerGraph(recipeGroup, workers, recipe):
    for workerGroup in recipeGroup:
        workers[workerGroup._v_name]=instantiateWorker(recipe, workerGroup)
    for workerGroup in recipeGroup:
        for plugGroup in workerGroup.plugs:
            plug=workers[workerGroup._v_name].getPlug(plugGroup._v_name)
            for connection in plugGroup.connections.iterrows():
                workers[connection['destinationWorker']].getSocket(connection['destinationSocket']).insert(plug)

def restoreResultsToWorkers(recipeGroup, workers, h5):
    for workerGroup in recipeGroup:
        for plugGroup in workerGroup.plugs:
            plug=workers[workerGroup._v_name].getPlug(plugGroup._v_name)
            try:
                resId = plugGroup._v_attrs.result
                resNode = h5.getNode("/results/"+resId)
                hash, uriType = DataContainer.parseId(resNode._v_title)
                if uriType==u'field':
                    result=loadField(h5, resNode)
                elif uriType==u'sample':
                    _logger.info("Trying to load sample data...")
                    result=loadSample(h5, resNode)
                    _logger.info("...successfully loaded.")
                else:
                    raise TypeError, "Unknown result uriType in <%s>"%resNode._v_title
                plug._result = result
            except (AttributeError, tables.NoSuchNodeError), e:
                _logger.info( "Exception: "+str(e) )

def loadField(h5, resNode):
    longname = unicode(h5.getNodeAttr(resNode, "longname"), 'utf-8')
    shortname = unicode(h5.getNodeAttr(resNode, "shortname"), 'utf-8')
    try:
        creator = unicode(h5.getNodeAttr(resNode, "creator"), 'utf-8')
        machine = unicode(h5.getNodeAttr(resNode, "machine"), 'utf-8')
    except:
        from pyphant.core.Helpers import emd52dict
        emd5dict = emd52dict(resNode._v_title)
        creator = emd5dict['creator']
        machine = emd5dict['machine']
    data = scipy.array(resNode.data.read())
    def loads(inputList):
        if type(inputList)==type([]):
            try:
                return map(lambda s: eval(s),inputList)
            except:
                return map(lambda s: unicode(s, 'utf-8'),inputList)
        else:
            return map(loads,inputList)
    if data.dtype.char == 'S':
        data = scipy.array(loads(data.tolist()))
    attributes = {}
    for key in resNode.data._v_attrs._v_attrnamesuser:
        attributes[key]=h5.getNodeAttr(resNode.data,key)
    try:
        error = scipy.array(resNode.error.read())
    except tables.NoSuchNodeError, e:
        error = None
    try:
        mask = scipy.array(resNode.mask.read())
    except tables.NoSuchNodeError, e:
        mask = None
    unit = eval(unicode(h5.getNodeAttr(resNode, "unit"), 'utf-8'))
    try:
        dimTable = resNode.dimensions
        dimensions = [loadField(h5, h5.getNode("/results/result_"+DataContainer.parseId(row['id'])[0]))
                      for row in dimTable.iterrows()]
    except tables.NoSuchNodeError, e:
        dimensions = DataContainer.INDEX
    result = DataContainer.FieldContainer(data, unit, error, mask,
                                          dimensions, longname, shortname,
                                          attributes)
    result.creator = creator
    result.machine = machine
    result.seal(resNode._v_title)
    return result

def loadSample(h5, resNode):
    result = DataContainer.SampleContainer.__new__(DataContainer.SampleContainer)
    result.longname = unicode(h5.getNodeAttr(resNode, "longname"), 'utf-8')
    result.shortname = unicode(h5.getNodeAttr(resNode, "shortname"), 'utf-8')
    result.creator = unicode(h5.getNodeAttr(resNode, "creator"), 'utf-8')
    result.machine = unicode(h5.getNodeAttr(resNode, "machine"), 'utf-8')
    result.attributes = {}
    for key in resNode._v_attrs._v_attrnamesuser:
        if key not in _reservedAttributes:
            result.attributes[key]=h5.getNodeAttr(resNode,key)
    columns = []
    for resId in h5.getNodeAttr(resNode,"columns"):
        nodename = "/results/"+resId
        hash, uriType = DataContainer.parseId(h5.getNodeAttr(nodename, "TITLE"))
        if uriType == 'sample':
            loader = loadSample
        elif uriType =='field':
            loader = loadField
        else:
            raise KeyError, "Unknown UriType %s in saving result %s." % (uriType, result.id)
        columns.append(loader(h5,h5.getNode(nodename)))
    result.columns=columns
    result.seal(resNode._v_title)
    return result

def loadExecutionOrders(h5):
    orders = []
    for orderGroup in h5.root.executionOrder:
        socketMapTable = orderGroup.input
        socketMap = dict([(row['socket'], row['data']) for row in socketMapTable.iterrows()])
        resultPlug = orderGroup._v_attrs.resultPlug
        orders.append((socketMap, resultPlug))
    return orders

def pruneResults(h5):
    h5.removeNode("/results", recursive=True)
    h5.createGroup("/", "results")



# -*- coding: utf-8 -*-

# Copyright (c) 2006-2008, Rectorate of the University of Freiburg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of the Freiburg Materials Research Center,
#   University of Freiburg nor the names of its contributors may be used to
#   endorse or promote products derived from this software without specific
#   prior written permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
This module provides the KnowledgeNode class which is used as an
HTTP communication channel between one local KnowledgeManager and
arbitrary many remote KnowledgeManagers. It comes with a RoutingHTTPServer
and an optional WebInterface.
"""

from __future__ import with_statement
from pyphant.core.RoutingHTTPServer import (RoutingHTTPServer,
                                            UnreachableError)
import sqlite3
from pyphant.core.Helpers import getPyphantPath
from pyphant.core.SQLiteWrapper import create_table
from time import time
from urllib2 import (urlopen, URLError, HTTPError)
from urllib import urlencode
import logging
from pyphant.core.KnowledgeManager import (DCNotFoundError, KnowledgeManager)
from pyphant.core.bottle import (request, send_file)
try:
    from json import (dumps, load, loads)
except ImportError:
    from simplejson import (dumps, load, loads)
from tempfile import (mkdtemp, mkstemp)
import os
import pyphant.core.bottle
import pkg_resources


class SkipError(Exception):
    pass


class RemoteError(Exception):
    pass


class RemoteKN(object):
    """
    This class represents a remote KnowledgeNode.
    """

    status_dict = {0: 'offline', 1: 'online', 2: 'disabled'}

    def __init__(self, host, port, status=1, timeout=300.0):
        """
        Arguments:
        - `host`: hostname
        - `port`: port
        - `status`: 0: offline, may get online after timeout
                    1: online, may get offline anytime
                    2: disabled, use enable() to enable
        - `timeout`: refresh interval for status lookup when offline
                     default: 5 min
        """
        self.host = host
        self.port = port
        self.url = "http://%s:%d/" % (host, port)
        self.timeout = timeout
        self.last_update = None
        self.uuid = None
        self._status = status
        self.logger = logging.getLogger('pyphant')
        self.update_status()

    def __eq__(self, other):
        if not isinstance(other, RemoteKN):
            return False
        else:
            return self.host == other.host and self.port == other.port

    def _get_status(self):
        return self.status_dict[self._status]
    status = property(_get_status)

    def enable(self):
        self._status = 0
        self.update_status()

    def disable(self):
        self.last_update = None
        self.uuid = None
        self._status = 2

    def update_status(self):
        if self._status == 2:
            return
        elif self.last_update == None or self._status == 1:
            self.connect()
        else:
            if time() - self.last_update > self.timeout:
                self.connect()

    def connect(self):
        stream = None
        try:
            try:
                stream = urlopen(self.url + 'uuid/', timeout=3.0)
            except TypeError:
                stream = urlopen(self.url + 'uuid/')
            line = stream.readline()
            if line.startswith('urn:uuid:'):
                self._status = 1
                self.uuid = line
            else:
                self._status = 0
                self.logger.error("Remote KM '%s' returned broken uuid: '%s'" \
                                  % (self.url, line))
        except (URLError, IOError, HTTPError):
            self._status = 0
            self.logger.warn("Remote KM '%s' is not responding." % self.url)
        finally:
            if stream != None:
                stream.close()
            self.last_update = time()

    def get_datacontainer_url(self, dc_id, skip):
        self.update_status()
        if self._status == 1:
            if self.uuid in skip:
                raise SkipError()
            else:
                try:
                    query = urlencode({'skip': dumps(skip), 'dc_id': dc_id})
                    url = '%sget_dc_url/?%s' % (self.url, query)
                    try:
                        stream = urlopen(url, timeout=60.0)
                    except TypeError:
                        stream = urlopen(url)
                    assert stream.headers.type == 'application/json'
                    answer = load(stream)
                    stream.close()
                    if answer['dc_url'] == None:
                        raise DCNotFoundError
                    assert len(answer['skip']) >= len(skip)
                    return answer['dc_url'], answer['skip']
                except (URLError, HTTPError, IOError, AssertionError):
                    raise UnreachableError()
        else:
            raise UnreachableError()


class KnowledgeNode(RoutingHTTPServer):
    """
    This class manages communication between one local and arbitrary many
    remote KM instances.
    """

    def __init__(self, local_km=None,
                 host=u'127.0.0.1', port=8080, start=False,
                 web_interface=False, dbase=u'default'):
        """
        Arguments:
        - `local_km`: Local KnowledgeManager instance to hook up to.
          If set to `None`, KnowledgeManager.getInstance() is used.
        - `host`: hostname to listen on
        - `port`: port to listen on
        - `start`: flag that indicates whether to start the server
        - `web_interface`: flag that indicates whether to enable
          the web interface. You can enable/disable it anytime by
          setting (KN instance).web_interface.enabled to `True`/`False`.
        - `dbase`: leave this to 'default', other values are allowed for
                   debug purposes
        """
        RoutingHTTPServer.__init__(self, host, port, start)
        if local_km == None:
            local_km = KnowledgeManager.getInstance()
        self.km = local_km
        self.remotes = []
        if dbase == u'default':
            self._dbase = os.path.join(getPyphantPath('sqlite3'),
                                       'kn_remotes.sqlite3')
        else:
            self._dbase = dbase
        self._restore_remotes()
        self._setup_routes()
        self._tempdir = mkdtemp(prefix='HDF5Wrap')
        tpl_path = pkg_resources.resource_filename('pyphant', 'web/templates')
        if not tpl_path in pyphant.core.bottle.TEMPLATE_PATH:
            pyphant.core.bottle.TEMPLATE_PATH.append(tpl_path)
        from pyphant.core.WebInterface import WebInterface
        self.web_interface = WebInterface(self, web_interface)
        self.km.node = self

    def _restore_remotes(self):
        """
        Loads remotes from dbase.
        """
        connection = sqlite3.connect(self._dbase)
        cursor = connection.cursor()
        try:
            columns = [('host', 'TEXT'), ('port', 'INT'), ('status', 'INT'),
                       ('', 'UNIQUE(host, port)')]
            create_table('kn_remotes', columns, cursor, ignore_exists=True)
            cursor.execute("SELECT * FROM kn_remotes")
            self.remotes = [RemoteKN(host, port, status) \
                            for host, port, status in cursor]
            connection.commit()
        finally:
            cursor.close()
            connection.close()

    def _setup_routes(self):
        self.app.add_route('/uuid/', self.get_uuid)
        self.app.add_route('/get_dc_url/', self.handle_datacontainer_url)
        self.app.add_route(r'/wrapped/:filename#..*\.hdf$#',
                           self.handle_wrapped)

    def stop(self):
        RoutingHTTPServer.stop(self)
        if not hasattr(self, '_tempdir'):
            return
        if os.path.isdir(self._tempdir):
            from shutil import rmtree
            try:
                rmtree(self._tempdir)
            except OSError:
                self.km.logger.warn("Could not delete '%s'." % self._tempdir)

    def register_remote(self, host, port):
        host = host.lower()
        port = int(port)
        connection = sqlite3.connect(self._dbase)
        cursor = connection.cursor()
        error = None
        try:
            try:
                cursor.execute("INSERT OR ABORT INTO kn_remotes "\
                               "(host, port, status) "\
                               "VALUES (?, ?, ?)", (host, port, 0))
                self.remotes.append(RemoteKN(host, port))
            except sqlite3.IntegrityError:
                error = RemoteError("Remote '%s:%d' already registered." \
                                    % (host, port))
            connection.commit()
        finally:
            cursor.close()
            connection.close()
        if not error is None:
            raise error

    def remove_remote(self, host, port):
        host = host.lower()
        port = int(port)
        dummy = RemoteKN(host, port, status=2)
        try:
            self.remotes.remove(dummy)
        except ValueError:
            raise RemoteError("Remote '%s:%d' is not registered." \
                              % (host, port))
        connection = sqlite3.connect(self._dbase)
        cursor = connection.cursor()
        try:
            cursor.execute("DELETE FROM kn_remotes "\
                           "WHERE host=? AND port=?", (host, port))
            connection.commit()
        finally:
            cursor.close()
            connection.close()

    def change_remote(self, host, port, status):
        host = host.lower()
        port = int(port)
        dummy = RemoteKN(host, port, status=2)
        for rem in self.remotes:
            if rem == dummy and rem._status != status:
                if status == 2:
                    rem.disable()
                else:
                    rem.enable()
                connection = sqlite3.connect(self._dbase)
                cursor = connection.cursor()
                try:
                    cursor.execute("UPDATE kn_remotes SET status=? "\
                                   "WHERE host=? AND port=?",
                                   (status, host, port))
                    connection.commit()
                finally:
                    cursor.close()
                    connection.close()
                return
        raise RemoteError("Remote '%s:%d' is not registered." \
                          % (host, port))

    def disable_remote(self, host, port):
        self.change_remote(host, port, 2)

    def enable_remote(self, host, port):
        self.change_remote(host, port, 0)

    def get_uuid(self):
        return self.km.uuid
    uuid = property(get_uuid)

    def get_datacontainer(self, dc_id):
        skip = [self.uuid]
        for remote in self.remotes:
            try:
                dc_url, skip = remote.get_datacontainer_url(dc_id, skip)
                self.km.registerURL(dc_url)
                return self.km.getDataContainer(dc_id)
            except (DCNotFoundError, UnreachableError, SkipError):
                pass
        raise DCNotFoundError()

    def handle_datacontainer_url(self):
        query = request.GET
        skip = loads(query['skip'])
        if self.uuid in skip:
            # This should not happen during normal operation
            self.km.logger.error(
                "KN '%s' has been queried although it is in the skip list.")
        else:
            skip.append(self.uuid)
        dc_id = query['dc_id']
        try:
            dc = self.km.getDataContainer(dc_id, try_remote=False)
            # Wrap data container in temporary HDF5 file
            osFileId, filename = mkstemp(suffix='.hdf',
                                         prefix='dcrequest-',
                                         dir=self._tempdir)
            os.close(osFileId)
            handler = self.km.getH5FileHandler(filename, 'w')
            with handler:
                handler.saveDataContainer(dc)
            dc_url = self.url + "wrapped/" + os.path.basename(filename)
        except DCNotFoundError:
            dc_url = None
            for remote in self.remotes:
                try:
                    dc_url, skip = remote.get_datacontainer_url(dc_id, skip)
                    break
                except (DCNotFoundError, UnreachableError, SkipError):
                    pass
        return {'skip': skip, 'dc_url': dc_url}

    def handle_wrapped(self, filename):
        send_file(filename, self._tempdir,
                  guessmime=False, mimetype='application/x-hdf')


def get_kn_autoport(ports, logger=None, *args, **kargs):
    """
    Returns a KnowledgeNode listening on the first free port in `ports`
    messages are logged to `logger` or stdout if `None`
    If no port is free, a socket.error (no. 98) is raised.
    """
    import socket

    def log(text):
        if logger is None:
            print text
        else:
            logger.warn(text)

    last_error = None
    for port in ports:
        try:
            kn = KnowledgeNode(port=port, *args, **kargs)
            return kn
        except socket.error, err:
            last_error = err
            try:
                #Python 2.6
                eno = err.errno
            except AttributeError:
                #Python 2.5
                eno = err.args[0]
            from errno import (EADDRINUSE, EACCES)
            if eno == EADDRINUSE:
                log("Port %d is already in use." % port)
            elif eno == EACCES:
                log("Port %d: Permission denied." % port)
            else:
                raise err
    raise last_error
